@@ -315,7 +315,7 @@ runAsm platform long_jumps e = go
           words = concatMap expand ops
           expand (SmallOp w) = [w]
           expand (LabelOp w) = expand (Op (e w))
-          expand (Op w) = if largeOps then largeArg platform w else [fromIntegral w]
+          expand (Op w) = if largeOps then largeArg platform (fromIntegral w) else [fromIntegral w]
 --        expand (LargeOp w) = largeArg platform w
       state $ \(st_i0,st_l0,st_p0) ->
         let st_i1 = addListToSS st_i0 (opcode : words)
@@ -358,7 +358,7 @@ inspectAsm platform long_jumps initial_offset
 largeArgInstr :: Word16 -> Word16
 largeArgInstr bci = bci_FLAG_LARGE_ARGS .|. bci
 
-largeArg :: Platform -> Word -> [Word16]
+largeArg :: Platform -> Word64 -> [Word16]
 largeArg platform w = case platformWordSize platform of
    PW8 -> [fromIntegral (w `shiftR` 48),
            fromIntegral (w `shiftR` 32),
@@ -395,7 +395,10 @@ assembleI platform i = case i of
   PUSH_BCO proto           -> do let ul_bco = assembleBCO platform proto
                                  p <- ioptr (liftM BCOPtrBCO ul_bco)
                                  emit bci_PUSH_G [Op p]
-  PUSH_ALTS proto pk
+  PUSH_ALTS proto          -> do let ul_bco = assembleBCO platform proto
+                                 p <- ioptr (liftM BCOPtrBCO ul_bco)
+                                 emit bci_PUSH_ALTS [Op p]
+  PUSH_ALTS_UNLIFTED proto pk
                            -> do let ul_bco = assembleBCO platform proto
                                  p <- ioptr (liftM BCOPtrBCO ul_bco)
                                  emit (push_alts pk) [Op p]
@@ -501,7 +504,8 @@ assembleI platform i = case i of
   SWIZZLE   stkoff n       -> emit bci_SWIZZLE [SmallOp stkoff, SmallOp n]
   JMP       l              -> emit bci_JMP [LabelOp l]
   ENTER                    -> emit bci_ENTER []
-  RETURN rep               -> emit (return_non_tuple rep) []
+  RETURN                   -> emit bci_RETURN []
+  RETURN_UNLIFTED rep      -> emit (return_unlifted rep) []
   RETURN_TUPLE             -> emit bci_RETURN_T []
   CCALL off m_addr i       -> do np <- addr m_addr
                                  emit bci_CCALL [SmallOp off, Op np, SmallOp i]
@@ -570,16 +574,16 @@ push_alts V16 = error "push_alts: vector"
 push_alts V32 = error "push_alts: vector"
 push_alts V64 = error "push_alts: vector"
 
-return_non_tuple :: ArgRep -> Word16
-return_non_tuple V   = bci_RETURN_V
-return_non_tuple P   = bci_RETURN_P
-return_non_tuple N   = bci_RETURN_N
-return_non_tuple L   = bci_RETURN_L
-return_non_tuple F   = bci_RETURN_F
-return_non_tuple D   = bci_RETURN_D
-return_non_tuple V16 = error "return_non_tuple: vector"
-return_non_tuple V32 = error "return_non_tuple: vector"
-return_non_tuple V64 = error "return_non_tuple: vector"
+return_unlifted :: ArgRep -> Word16
+return_unlifted V   = bci_RETURN_V
+return_unlifted P   = bci_RETURN_P
+return_unlifted N   = bci_RETURN_N
+return_unlifted L   = bci_RETURN_L
+return_unlifted F   = bci_RETURN_F
+return_unlifted D   = bci_RETURN_D
+return_unlifted V16 = error "return_unlifted: vector"
+return_unlifted V32 = error "return_unlifted: vector"
+return_unlifted V64 = error "return_unlifted: vector"
 
 {-
   we can only handle up to a fixed number of words on the stack,
