@@ -32,6 +32,10 @@ module GHC.Platform
    , PlatformMisc(..)
    , SseVersion (..)
    , BmiVersion (..)
+   , wordAlignment
+   -- * SSE and AVX
+   , isSseEnabled
+   , isSse2Enabled
    -- * Platform constants
    , PlatformConstants(..)
    , lookupPlatformConstants
@@ -50,6 +54,7 @@ import GHC.Read
 import GHC.ByteOrder (ByteOrder(..))
 import GHC.Platform.Constants
 import GHC.Platform.ArchOS
+import GHC.Types.Basic (Alignment, alignmentOf)
 import GHC.Utils.Panic.Plain
 
 import Data.Word
@@ -73,11 +78,49 @@ data Platform = Platform
    , platformTablesNextToCode         :: !Bool
       -- ^ Determines whether we will be compiling info tables that reside just
       --   before the entry code, or with an indirection to the entry code. See
-      --   TABLES_NEXT_TO_CODE in includes/rts/storage/InfoTables.h.
+      --   TABLES_NEXT_TO_CODE in rts/include/rts/storage/InfoTables.h.
+   , platformHasLibm                  :: !Bool
+      -- ^ Some platforms require that we explicitly link against @libm@ if any
+      -- math-y things are used (which we assume to include all programs). See
+      -- #14022.
+
    , platform_constants               :: !(Maybe PlatformConstants)
       -- ^ Constants such as structure offsets, type sizes, etc.
    }
-   deriving (Read, Show, Eq)
+   deriving (Read, Show, Eq, Ord)
+
+wordAlignment :: Platform -> Alignment
+wordAlignment platform = alignmentOf (platformWordSizeInBytes platform)
+
+-- -----------------------------------------------------------------------------
+-- SSE and AVX
+
+-- TODO: Instead of using a separate predicate (i.e. isSse2Enabled) to
+-- check if SSE is enabled, we might have x86-64 imply the -msse2
+-- flag.
+
+isSseEnabled :: Platform -> Bool
+isSseEnabled platform = case platformArch platform of
+    ArchX86_64 -> True
+    ArchX86    -> True
+    _          -> False
+
+isSse2Enabled :: Platform -> Bool
+isSse2Enabled platform = case platformArch platform of
+  -- We assume  SSE1 and SSE2 operations are available on both
+  -- x86 and x86_64. Historically we didn't default to SSE2 and
+  -- SSE1 on x86, which results in defacto nondeterminism for how
+  -- rounding behaves in the associated x87 floating point instructions
+  -- because variations in the spill/fpu stack placement of arguments for
+  -- operations would change the precision and final result of what
+  -- would otherwise be the same expressions with respect to single or
+  -- double precision IEEE floating point computations.
+    ArchX86_64 -> True
+    ArchX86    -> True
+    _          -> False
+
+-- -----------------------------------------------------------------------------
+-- Platform Constants
 
 platformConstants :: Platform -> PlatformConstants
 platformConstants platform = case platform_constants platform of
@@ -93,6 +136,7 @@ genericPlatform = Platform
    , platformHasGnuNonexecStack      = False
    , platformHasIdentDirective       = False
    , platformHasSubsectionsViaSymbols= False
+   , platformHasLibm                 = False
    , platformIsCrossCompiling        = False
    , platformLeadingUnderscore       = False
    , platformTablesNextToCode        = True
@@ -162,6 +206,7 @@ osElfTarget OSHaiku     = True
 osElfTarget OSQNXNTO    = False
 osElfTarget OSAIX       = False
 osElfTarget OSHurd      = True
+osElfTarget OSWasi      = False
 osElfTarget OSUnknown   = False
  -- Defaulting to False is safe; it means don't rely on any
  -- ELF-specific functionality.  It is important to have a default for
@@ -218,6 +263,7 @@ platformCConvNeedsExtension platform = case platformArch platform of
   ArchPPC_64 _ -> True
   ArchS390X    -> True
   ArchRISCV64  -> True
+  ArchLoongArch64 -> True
   ArchAArch64
       -- Apple's AArch64 ABI requires that the caller sign-extend
       -- small integer arguments. See
@@ -253,10 +299,7 @@ data PlatformMisc = PlatformMisc
   { -- TODO Recalculate string from richer info?
     platformMisc_targetPlatformString :: String
   , platformMisc_ghcWithInterpreter   :: Bool
-  , platformMisc_ghcWithSMP           :: Bool
-  , platformMisc_ghcRTSWays           :: String
   , platformMisc_libFFI               :: Bool
-  , platformMisc_ghcRtsWithLibdw      :: Bool
   , platformMisc_llvmTarget           :: String
   }
 
