@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP #-}
+
 
 -----------------------------------------------------------------------------
 --
@@ -12,7 +12,8 @@ module GHC.StgToCmm.Utils (
         emitDataLits, emitRODataLits,
         emitDataCon,
         emitRtsCall, emitRtsCallWithResult, emitRtsCallGen,
-        assignTemp,
+        emitBarf,
+        assignTemp, newTemp,
 
         newUnboxedTupleRegs,
 
@@ -45,14 +46,12 @@ module GHC.StgToCmm.Utils (
         convertInfoProvMap, cmmInfoTableToInfoProvEnt
   ) where
 
-#include "HsVersions.h"
-
-import GHC.Prelude
+import GHC.Prelude hiding ( head, init, last, tail )
 
 import GHC.Platform
 import GHC.StgToCmm.Monad
 import GHC.StgToCmm.Closure
-import GHC.StgToCmm.Lit (mkSimpleLit)
+import GHC.StgToCmm.Lit (mkSimpleLit, newStringCLit)
 import GHC.Cmm
 import GHC.Cmm.BlockId
 import GHC.Cmm.Graph as CmmGraph
@@ -72,10 +71,10 @@ import GHC.Types.Literal
 import GHC.Data.Graph.Directed
 import GHC.Utils.Misc
 import GHC.Types.Unique
-import GHC.Driver.Session
 import GHC.Data.FastString
 import GHC.Utils.Outputable
 import GHC.Utils.Panic
+import GHC.Utils.Panic.Plain
 import GHC.Types.RepType
 import GHC.Types.CostCentre
 import GHC.Types.IPE
@@ -85,12 +84,12 @@ import Data.List (sortBy)
 import Data.Ord
 import GHC.Types.Unique.Map
 import Data.Maybe
-import GHC.Driver.Ppr
 import qualified Data.List.NonEmpty as NE
 import GHC.Core.DataCon
 import GHC.Types.Unique.FM
 import GHC.Data.Maybe
 import Control.Monad
+import qualified Data.Map.Strict as Map
 
 --------------------------------------------------------------------------
 --
@@ -99,7 +98,7 @@ import Control.Monad
 --------------------------------------------------------------------------
 
 addToMemLbl :: CmmType -> CLabel -> Int -> CmmAGraph
-addToMemLbl rep lbl n = addToMem rep (CmmLit (CmmLabel lbl)) n
+addToMemLbl rep lbl = addToMem rep (CmmLit (CmmLabel lbl))
 
 addToMemLblE :: CmmType -> CLabel -> CmmExpr -> CmmAGraph
 addToMemLblE rep lbl = addToMemE rep (CmmLit (CmmLabel lbl))
@@ -159,13 +158,17 @@ tagToClosure platform tycon tag
 --
 -------------------------------------------------------------------------
 
+emitBarf :: String -> FCode ()
+emitBarf msg = do
+  strLbl <- newStringCLit msg
+  emitRtsCall rtsUnitId (fsLit "barf") [(CmmLit strLbl,AddrHint)] False
+
 emitRtsCall :: UnitId -> FastString -> [(CmmExpr,ForeignHint)] -> Bool -> FCode ()
-emitRtsCall pkg fun args safe = emitRtsCallGen [] (mkCmmCodeLabel pkg fun) args safe
+emitRtsCall pkg fun = emitRtsCallGen [] (mkCmmCodeLabel pkg fun)
 
 emitRtsCallWithResult :: LocalReg -> ForeignHint -> UnitId -> FastString
         -> [(CmmExpr,ForeignHint)] -> Bool -> FCode ()
-emitRtsCallWithResult res hint pkg fun args safe
-   = emitRtsCallGen [(res,hint)] (mkCmmCodeLabel pkg fun) args safe
+emitRtsCallWithResult res hint pkg = emitRtsCallGen [(res,hint)] . mkCmmCodeLabel pkg
 
 -- Make a call to an RTS C procedure
 emitRtsCallGen
@@ -175,7 +178,7 @@ emitRtsCallGen
    -> Bool -- True <=> CmmSafe call
    -> FCode ()
 emitRtsCallGen res lbl args safe
-  = do { platform <- targetPlatform <$> getDynFlags
+  = do { platform <- getPlatform
        ; updfr_off <- getUpdFrameOff
        ; let (caller_save, caller_load) = callerSaveVolatileRegs platform
        ; emit caller_save
@@ -202,7 +205,7 @@ emitRtsCallGen res lbl args safe
 -- Here we generate the sequence of saves/restores required around a
 -- foreign call instruction.
 
--- TODO: reconcile with includes/Regs.h
+-- TODO: reconcile with rts/include/Regs.h
 --  * Regs.h claims that BaseReg should be saved last and loaded first
 --    * This might not have been tickled before since BaseReg is callee save
 --  * Regs.h saves SparkHd, ParkT1, SparkBase and SparkLim
@@ -210,7 +213,7 @@ emitRtsCallGen res lbl args safe
 -- This code isn't actually used right now, because callerSaves
 -- only ever returns true in the current universe for registers NOT in
 -- system_regs (just do a grep for CALLER_SAVES in
--- includes/stg/MachRegs.h).  It's all one giant no-op, and for
+-- rts/include/stg/MachRegs.h).  It's all one giant no-op, and for
 -- good reason: having to save system registers on every foreign call
 -- would be very expensive, so we avoid assigning them to those
 -- registers when we add support for an architecture.
@@ -263,7 +266,7 @@ emitRODataLits lbl lits = emitDecl (mkRODataLits lbl lits)
 
 emitDataCon :: CLabel -> CmmInfoTable -> CostCentreStack -> [CmmLit] -> FCode ()
 emitDataCon lbl itbl ccs payload =
-  emitDecl (CmmData (Section Data lbl) (CmmStatics lbl itbl ccs payload))
+  emitDecl (CmmData (Section Data lbl) (CmmStatics lbl itbl ccs payload []))
 
 -------------------------------------------------------------------------
 --
@@ -291,19 +294,17 @@ newUnboxedTupleRegs :: Type -> FCode ([LocalReg], [ForeignHint])
 -- the Sequel.  If the Sequel is a join point, using the
 -- regs it wants will save later assignments.
 newUnboxedTupleRegs res_ty
-  = ASSERT( isUnboxedTupleType res_ty )
+  = assert (isUnboxedTupleType res_ty) $
     do  { platform <- getPlatform
         ; sequel <- getSequel
         ; regs <- choose_regs platform sequel
-        ; ASSERT( regs `equalLength` reps )
-          return (regs, map primRepForeignHint reps) }
+        ; massert (regs `equalLength` reps)
+        ; return (regs, map primRepForeignHint reps) }
   where
     reps = typePrimRep res_ty
     choose_regs _ (AssignTo regs _) = return regs
     choose_regs platform _          = mapM (newTemp . primRepCmmType platform) reps
 
-
-
 -------------------------------------------------------------------------
 --      emitMultiAssign
 -------------------------------------------------------------------------
@@ -327,7 +328,7 @@ emitMultiAssign []    []    = return ()
 emitMultiAssign [reg] [rhs] = emitAssign (CmmLocal reg) rhs
 emitMultiAssign regs rhss   = do
   platform <- getPlatform
-  ASSERT2( equalLength regs rhss, ppr regs $$ pdoc platform rhss )
+  assertPpr (equalLength regs rhss) (ppr regs $$ pdoc platform rhss) $
     unscramble platform ([1..] `zip` (regs `zip` rhss))
 
 unscramble :: Platform -> [Vrtx] -> FCode ()
@@ -415,7 +416,7 @@ mk_discrete_switch :: Bool -- ^ Use signed comparisons
 -- SINGLETON TAG RANGE: no case analysis to do
 mk_discrete_switch _ _tag_expr [(tag, lbl)] _ (lo_tag, hi_tag)
   | lo_tag == hi_tag
-  = ASSERT( tag == lo_tag )
+  = assert (tag == lo_tag) $
     mkBranch lbl
 
 -- SINGLETON BRANCH, NO DEFAULT: no case analysis to do
@@ -447,8 +448,8 @@ emitCmmLitSwitch :: CmmExpr                    -- Tag to switch on
                -> [(Literal, CmmAGraphScoped)] -- Tagged branches
                -> CmmAGraphScoped              -- Default branch (always)
                -> FCode ()                     -- Emit the code
-emitCmmLitSwitch _scrut []       deflt = emit $ fst deflt
-emitCmmLitSwitch scrut  branches deflt = do
+emitCmmLitSwitch _scrut [] deflt = emit $ fst deflt
+emitCmmLitSwitch scrut branches@(branch:_) deflt = do
     scrut' <- assignTemp' scrut
     join_lbl <- newBlockId
     deflt_lbl <- label_code join_lbl deflt
@@ -459,12 +460,19 @@ emitCmmLitSwitch scrut  branches deflt = do
         rep = typeWidth cmm_ty
 
     -- We find the necessary type information in the literals in the branches
-    let signed = case head branches of
-                    (LitNumber nt _, _) -> litNumIsSigned nt
-                    _ -> False
-
-    let range | signed    = (platformMinInt platform, platformMaxInt platform)
-              | otherwise = (0, platformMaxWord platform)
+    let (signed,range) = case branch of
+          (LitNumber nt _, _) -> (signed,range)
+            where
+              signed = litNumIsSigned nt
+              range  = case litNumRange platform nt of
+                        (Just mi, Just ma) -> (mi,ma)
+                                              -- unbounded literals (Natural and
+                                              -- Integer) must have been
+                                              -- lowered at this point
+                        partial_bounds     -> pprPanic "Unexpected unbounded literal range"
+                                                       (ppr partial_bounds)
+               -- assuming native word range
+          _ -> (False, (0, platformMaxWord platform))
 
     if isFloatType cmm_ty
     then emit =<< mk_float_switch rep scrut' deflt_lbl noBound branches_lbls
@@ -596,15 +604,15 @@ cmmInfoTableToInfoProvEnt this_mod cmit =
     in InfoProvEnt cl cn "" this_mod Nothing
 
 -- | Convert source information collected about identifiers in 'GHC.STG.Debug'
--- to entries suitable for placing into the info table provenenance table.
-convertInfoProvMap :: DynFlags -> [CmmInfoTable] -> Module -> InfoTableProvMap -> [InfoProvEnt]
-convertInfoProvMap dflags defns this_mod (InfoTableProvMap (UniqMap dcenv) denv) =
+-- to entries suitable for placing into the info table provenance table.
+convertInfoProvMap :: [CmmInfoTable] -> Module -> InfoTableProvMap -> [InfoProvEnt]
+convertInfoProvMap defns this_mod (InfoTableProvMap (UniqMap dcenv) denv infoTableToSourceLocationMap) =
   map (\cmit ->
     let cl = cit_lbl cmit
         cn  = rtsClosureType (cit_rep cmit)
 
         tyString :: Outputable a => a -> String
-        tyString t = showPpr dflags t
+        tyString = renderWithContext defaultSDocContext . ppr
 
         lookupClosureMap :: Maybe InfoProvEnt
         lookupClosureMap = case hasHaskellName cl >>= lookupUniqMap denv of
@@ -614,12 +622,20 @@ convertInfoProvMap dflags defns this_mod (InfoTableProvMap (UniqMap dcenv) denv)
         lookupDataConMap = do
             UsageSite _ n <- hasIdLabelInfo cl >>= getConInfoTableLocation
             -- This is a bit grimy, relies on the DataCon and Name having the same Unique, which they do
-            (dc, ns) <- (hasHaskellName cl >>= lookupUFM_Directly dcenv . getUnique)
+            (dc, ns) <- hasHaskellName cl >>= lookupUFM_Directly dcenv . getUnique
             -- Lookup is linear but lists will be small (< 100)
             return $ InfoProvEnt cl cn (tyString (dataConTyCon dc)) this_mod (join $ lookup n (NE.toList ns))
 
+        lookupInfoTableToSourceLocation = do
+            sourceNote <- Map.lookup (cit_lbl cmit) infoTableToSourceLocationMap
+            return $ InfoProvEnt cl cn "" this_mod sourceNote
+
         -- This catches things like prim closure types and anything else which doesn't have a
         -- source location
         simpleFallback = cmmInfoTableToInfoProvEnt this_mod cmit
 
-    in fromMaybe simpleFallback (lookupDataConMap `firstJust` lookupClosureMap)) defns
+  in
+    if (isStackRep . cit_rep) cmit then
+      fromMaybe simpleFallback lookupInfoTableToSourceLocation
+    else
+      fromMaybe simpleFallback (lookupDataConMap `firstJust` lookupClosureMap)) defns
