@@ -67,7 +67,9 @@ module GHC.Types.Literal
 import GHC.Prelude
 
 import GHC.Builtin.Types.Prim
-import GHC.Core.Type
+import GHC.Core.Type( Type, RuntimeRepType, mkForAllTy, mkTyVarTy, typeOrConstraintKind )
+import GHC.Core.TyCo.Compare( nonDetCmpType )
+import GHC.Types.Var
 import GHC.Utils.Outputable
 import GHC.Data.FastString
 import GHC.Types.Basic
@@ -82,7 +84,7 @@ import Data.Int
 import Data.Word
 import Data.Char
 import Data.Data ( Data )
-import GHC.Exts
+import GHC.Exts( isTrue#, dataToTag#, (<#) )
 import Numeric ( fromRat )
 
 {-
@@ -131,15 +133,14 @@ data Literal
                                 -- that can be represented as a Literal. Create
                                 -- with 'nullAddrLit'
 
-  | LitRubbish Type             -- ^ A nonsense value of the given
-                                -- representation. See Note [Rubbish literals].
-                                --
-                                -- The Type argument, rr, is of kind RuntimeRep.
-                                -- The type of the literal is forall (a:TYPE rr). a
-                                --
-                                -- INVARIANT: the Type has no free variables
-                                --    and so substitution etc can ignore it
-                                --
+  | LitRubbish                  -- ^ A nonsense value; See Note [Rubbish literals].
+      TypeOrConstraint          -- t_or_c: whether this is a type or a constraint
+      RuntimeRepType            -- rr: a type of kind RuntimeRep
+      -- The type of the literal is forall (a:TYPE rr). a
+      --                         or forall (a:CONSTRAINT rr). a
+      --
+      -- INVARIANT: the Type has no free variables
+      --    and so substitution etc can ignore it
 
   | LitFloat   Rational         -- ^ @Float#@. Create with 'mkLitFloat'
   | LitDouble  Rational         -- ^ @Double#@. Create with 'mkLitDouble'
@@ -267,7 +268,7 @@ instance Binary Literal where
         = do putByte bh 6
              put_ bh nt
              put_ bh i
-    put_ _ (LitRubbish b) = pprPanic "Binary LitRubbish" (ppr b)
+    put_ _ lit@(LitRubbish {}) = pprPanic "Binary LitRubbish" (ppr lit)
      -- We use IfaceLitRubbish; see Note [Rubbish literals], item (6)
 
     get bh = do
@@ -297,6 +298,7 @@ instance Binary Literal where
                     return (LitNumber nt i)
               _ -> pprPanic "Binary:Literal" (int (fromIntegral h))
 
+
 instance Outputable Literal where
     ppr = pprLiteral id
 
@@ -578,7 +580,7 @@ mkLitChar = LitChar
 mkLitString :: String -> Literal
 -- stored UTF-8 encoded
 mkLitString [] = LitString mempty
-mkLitString s  = LitString (utf8EncodeString s)
+mkLitString s  = LitString (utf8EncodeByteString s)
 
 mkLitBigNat :: Integer -> Literal
 mkLitBigNat x = assertPpr (x >= 0) (integer x)
@@ -850,10 +852,10 @@ literalType (LitNumber lt _)  = case lt of
    LitNumWord64  -> word64PrimTy
 
 -- LitRubbish: see Note [Rubbish literals]
-literalType (LitRubbish rep)
-  = mkForAllTy a Inferred (mkTyVarTy a)
+literalType (LitRubbish torc rep)
+  = mkForAllTy (Bndr a Inferred) (mkTyVarTy a)
   where
-    a = mkTemplateKindVar (mkTYPEapp rep)
+    a = mkTemplateKindVar (typeOrConstraintKind torc rep)
 
 {-
         Comparison
@@ -869,7 +871,8 @@ cmpLit (LitDouble    a)     (LitDouble     b)     = a `compare` b
 cmpLit (LitLabel     a _ _) (LitLabel      b _ _) = a `lexicalCompareFS` b
 cmpLit (LitNumber nt1 a)    (LitNumber nt2  b)
   = (nt1 `compare` nt2) `mappend` (a `compare` b)
-cmpLit (LitRubbish b1)      (LitRubbish b2)       = b1 `nonDetCmpType` b2
+cmpLit (LitRubbish tc1 b1)  (LitRubbish tc2 b2)  = (tc1 `compare` tc2) `mappend`
+                                                   (b1 `nonDetCmpType` b2)
 cmpLit lit1 lit2
   | isTrue# (dataToTag# lit1 <# dataToTag# lit2) = LT
   | otherwise                                    = GT
@@ -903,9 +906,13 @@ pprLiteral add_par (LitLabel l mb fod) =
     add_par (text "__label" <+> b <+> ppr fod)
     where b = case mb of
               Nothing -> pprHsString l
-              Just x  -> doubleQuotes (text (unpackFS l ++ '@':show x))
-pprLiteral _       (LitRubbish rep)
-  = text "RUBBISH" <> parens (ppr rep)
+              Just x  -> doubleQuotes (ftext l <> text ('@':show x))
+pprLiteral _       (LitRubbish torc rep)
+  = text "RUBBISH" <> pp_tc <> parens (ppr rep)
+  where
+  pp_tc = case torc of
+           TypeLike       -> empty
+           ConstraintLike -> text "[c]"
 
 {-
 Note [Printing of literals in Core]
@@ -964,7 +971,7 @@ data type. Here are the moving parts:
    an IR feature.
 
 2. Core: 'LitRubbish' carries a `Type` of kind RuntimeRep,
-   describing the runtime representaion of the literal (is it a
+   describing the runtime representation of the literal (is it a
    pointer, an unboxed Double#, or whatever).
 
    We have it that `RUBBISH[rr]` has type `forall (a :: TYPE rr). a`.
@@ -1004,16 +1011,16 @@ data type. Here are the moving parts:
    all boxed to the host GC anyway.
 
 6. IfaceSyn: `Literal` is part of `IfaceSyn`, but `Type` really isn't.  So in
-   the passage from Core to Iface I put LitRubbish into its owns IfaceExpr data
+   the passage from Core to Iface we put LitRubbish into its own IfaceExpr data
    constructor, IfaceLitRubbish. The remaining constructors of Literal are
    fine as IfaceSyn.
 
 Wrinkles
 
 a) Why do we put the `Type` (of kind RuntimeRep) inside the literal?  Could
-   we not instead /apply/ the literal to that RuntimeRep?  Alas no, becuase
+   we not instead /apply/ the literal to that RuntimeRep?  Alas no, because
    then LitRubbish :: forall (rr::RuntimeRep) (a::TYPE rr). a
-   and that's am ill-formed type because its kind is `TYPE rr`, which escapes
+   and that's an ill-formed type because its kind is `TYPE rr`, which escapes
    the binding site of `rr`. Annoying.
 
 b) A rubbish literal is not bottom, and replies True to exprOkForSpeculation.
@@ -1043,9 +1050,9 @@ Suppose there is a bug in GHC, and a rubbish value is used after all. That is
 undefined behavior, of course, but let us list a few examples for failure modes:
 
  a) For an value of unboxed numeric type like `Int#`, we just use a silly
-    value like 42#. The error might propoagate indefinitely, hence we better
+    value like 42#. The error might propagate indefinitely, hence we better
     pick a rather unique literal. Same for Word, Floats, Char and VecRep.
- b) For AddrRep (like String lits), we mit a null pointer, resulting in a
+ b) For AddrRep (like String lits), we emit a null pointer, resulting in a
     definitive segfault when accessed.
  c) For boxed values, unlifted or not, we use a pointer to a fixed closure,
     like `()`, so that the GC has a pointer to follow.
