@@ -1,22 +1,23 @@
 module Rules.Generate (
     isGeneratedCmmFile, compilerDependencies, generatePackageCode,
     generateRules, copyRules, generatedDependencies,
-    ghcPrimDependencies
+    ghcPrimDependencies,
+    templateRules
     ) where
 
-import Data.Foldable (for_)
-
+import qualified Data.Set as Set
 import Base
 import qualified Context
 import Expression
-import Hadrian.Oracles.TextFile (lookupValueOrError)
+import Hadrian.Oracles.TextFile (lookupSystemConfig)
 import Oracles.Flag
 import Oracles.ModuleFiles
 import Oracles.Setting
+import Hadrian.Haskell.Cabal.Type (PackageData(version))
+import Hadrian.Oracles.Cabal (readPackageData)
 import Packages
 import Rules.Libffi
 import Settings
-import Settings.Builders.DeriveConstants (deriveConstantsPairs)
 import Target
 import Utilities
 
@@ -39,21 +40,41 @@ ghcPrimDependencies = do
     path  <- expr $ buildPath (vanillaContext stage ghcPrim)
     return [path -/- "GHC/Prim.hs", path -/- "GHC/PrimopWrappers.hs"]
 
-derivedConstantsFiles :: [FilePath]
-derivedConstantsFiles =
-    [ "DerivedConstants.h"
-    ]
+rtsDependencies :: Expr [FilePath]
+rtsDependencies = do
+    stage   <- getStage
+    rtsPath <- expr (rtsBuildPath stage)
+    jsTarget <- expr isJsTarget
+    useSystemFfi <- expr (flag UseSystemFfi)
+
+    let -- headers common to native and JS RTS
+        common_headers =
+            [ "ghcautoconf.h", "ghcplatform.h"
+            , "DerivedConstants.h"
+            ]
+        -- headers specific to the native RTS
+        native_headers =
+            [ "rts" -/- "EventTypes.h"
+            , "rts" -/- "EventLogConstants.h"
+            ]
+            ++ (if useSystemFfi then [] else libffiHeaderFiles)
+        headers
+          | jsTarget  = common_headers
+          | otherwise = common_headers ++ native_headers
+    pure $ ((rtsPath -/- "include") -/-) <$> headers
+
+genapplyDependencies :: Expr [FilePath]
+genapplyDependencies = do
+    stage   <- getStage
+    rtsPath <- expr (rtsBuildPath $ succStage stage)
+    ((stage /= Stage3) ?) $ pure $ ((rtsPath -/- "include") -/-) <$>
+        [ "ghcautoconf.h", "ghcplatform.h" ]
 
 compilerDependencies :: Expr [FilePath]
 compilerDependencies = do
     stage   <- getStage
     ghcPath <- expr $ buildPath (vanillaContext stage compiler)
-    rtsPath <- expr (rtsBuildPath stage)
-    libDir <- expr $ stageLibPath stage
-    useSystemFfi <- expr (flag UseSystemFfi)
-    mconcat [ return $ (libDir -/-) <$> derivedConstantsFiles
-            , notStage0 ? not useSystemFfi ? return ((rtsPath -/-) <$> libffiHeaderFiles)
-            , return $ fmap (ghcPath -/-)
+    pure $ (ghcPath -/-) <$>
                   [ "primop-can-fail.hs-incl"
                   , "primop-code-size.hs-incl"
                   , "primop-commutable.hs-incl"
@@ -71,21 +92,16 @@ compilerDependencies = do
                   , "primop-vector-uniques.hs-incl"
                   , "primop-docs.hs-incl"
                   , "GHC/Platform/Constants.hs"
-                  ] ]
+                  , "GHC/Settings/Config.hs"
+                  ]
 
 generatedDependencies :: Expr [FilePath]
 generatedDependencies = do
-    stage    <- getStage
-    rtsPath  <- expr (rtsBuildPath stage)
-    useSystemFfi <- expr (flag UseSystemFfi)
-    includes <- expr $ includesDependencies stage
-    libDir <- expr $ stageLibPath stage
     mconcat [ package compiler ? compilerDependencies
             , package ghcPrim  ? ghcPrimDependencies
-            , package rts      ? return ((if useSystemFfi then pure [] else fmap (rtsPath -/-) libffiHeaderFiles)
-                ++ includes
-                ++ ((libDir -/-) <$> derivedConstantsFiles))
-            , stage0 ? return includes ]
+            , package rts      ? rtsDependencies
+            , package genapply ? genapplyDependencies
+            ]
 
 generate :: FilePath -> Context -> Expr String -> Action ()
 generate file context expr = do
@@ -94,7 +110,7 @@ generate file context expr = do
     putSuccess $ "| Successfully generated " ++ file ++ "."
 
 generatePackageCode :: Context -> Rules ()
-generatePackageCode context@(Context stage pkg _) = do
+generatePackageCode context@(Context stage pkg _ _) = do
     root <- buildRootRules
     let dir         = buildDir context
         generated f = (root -/- dir -/- "**/*.hs") ?== f && not ("//autogen/*" ?== f)
@@ -102,7 +118,9 @@ generatePackageCode context@(Context stage pkg _) = do
     generated ?> \file -> do
         let unpack = fromMaybe . error $ "No generator for " ++ file ++ "."
         (src, builder) <- unpack <$> findGenerator context file
-        need [src]
+        -- Make sure we have configured the package before running the builder
+        pkg_setup <- pkgSetupConfigFile context
+        need [src, pkg_setup]
         build $ target context builder [src] [file]
         let boot = src -<.> "hs-boot"
         whenM (doesFileExist boot) $ do
@@ -124,25 +142,28 @@ generatePackageCode context@(Context stage pkg _) = do
 
     when (pkg == compiler) $ do
         root -/- primopsTxt stage %> \file -> do
-            includes <- includesDependencies stage
-            need $ [primopsSource] ++ includes
+            need $ [primopsSource]
             build $ target context HsCpp [primopsSource] [file]
 
     when (pkg == rts) $ do
         root -/- "**" -/- dir -/- "cmm/AutoApply.cmm" %> \file ->
             build $ target context GenApply [] [file]
-        -- TODO: This should be fixed properly, e.g. generated here on demand.
-        (root -/- "**" -/- dir -/- "DerivedConstants.h") <~ stageLibPath stage
-        (root -/- "**" -/- dir -/- "ghcautoconf.h") <~ stageLibPath stage
-        (root -/- "**" -/- dir -/- "ghcplatform.h") <~ stageLibPath stage
-        (root -/- "**" -/- dir -/- "ghcversion.h") <~ stageLibPath stage
- where
-    pattern <~ mdir = pattern %> \file -> do
-        dir <- mdir
-        copyFile (dir -/- takeFileName file) file
+        let go gen file = generate file (semiEmptyTarget stage) gen
+        root -/- "**" -/- dir -/- "include/ghcautoconf.h" %> go generateGhcAutoconfH
+        root -/- "**" -/- dir -/- "include/ghcplatform.h" %> go generateGhcPlatformH
+        root -/- "**" -/- dir -/- "include/DerivedConstants.h" %> genPlatformConstantsHeader context
+        root -/- "**" -/- dir -/- "include/rts/EventLogConstants.h" %> genEventTypes "--event-types-defines"
+        root -/- "**" -/- dir -/- "include/rts/EventTypes.h" %> genEventTypes "--event-types-array"
+
+genEventTypes :: String -> FilePath -> Action ()
+genEventTypes flag file = do
+    need ["rts" -/- "gen_event_types.py"]
+    runBuilder Python
+      ["rts" -/- "gen_event_types.py", flag, file]
+      [] []
 
 genPrimopCode :: Context -> FilePath -> Action ()
-genPrimopCode context@(Context stage _pkg _) file = do
+genPrimopCode context@(Context stage _pkg _ _) file = do
     root <- buildRoot
     need [root -/- primopsTxt stage]
     build $ target context GenPrimopCode [root -/- primopsTxt stage] [file]
@@ -150,12 +171,24 @@ genPrimopCode context@(Context stage _pkg _) file = do
 genPlatformConstantsType :: Context -> FilePath -> Action ()
 genPlatformConstantsType context file = do
     withTempDir $ \tmpdir ->
-      build $ target context DeriveConstants [] [file,"--gen-haskell-type",tmpdir]
+      build $ target context DeriveConstants [] [file,tmpdir]
+
+genPlatformConstantsHeader :: Context -> FilePath -> Action ()
+genPlatformConstantsHeader context file = do
+    -- N.B. deriveConstants needs to compile programs which #include
+    -- PosixSource.h, which #include's ghcplatform.h. Fixes #18290.
+    let prefix = takeDirectory file
+    need
+        [ prefix -/- "ghcplatform.h"
+        , prefix -/- "ghcautoconf.h"
+        ]
+    withTempDir $ \tmpdir -> build $
+        target context DeriveConstants [] [file, tmpdir]
 
 copyRules :: Rules ()
 copyRules = do
     root <- buildRootRules
-    forM_ [Stage0 ..] $ \stage -> do
+    forM_ allStages $ \stage -> do
         let prefix = root -/- stageString stage -/- "lib"
 
             infixl 1 <~
@@ -172,32 +205,23 @@ copyRules = do
         prefix -/- "html/**"           <~ return "utils/haddock/haddock-api/resources"
         prefix -/- "latex/**"          <~ return "utils/haddock/haddock-api/resources"
 
+        forM_ [Inplace, Final] $ \iplace ->
+          root -/- relativePackageDbPath (PackageDbLoc stage iplace) -/- systemCxxStdLibConf %> \file -> do
+            copyFile ("mk" -/- "system-cxx-std-lib-1.0.conf") file
+
 generateRules :: Rules ()
 generateRules = do
     root <- buildRootRules
 
     (root -/- "ghc-stage1") <~+ ghcWrapper Stage1
     (root -/- "ghc-stage2") <~+ ghcWrapper Stage2
+    (root -/- "ghc-stage3") <~+ ghcWrapper Stage3
 
-    forM_ [Stage0 ..] $ \stage -> do
+    forM_ allStages $ \stage -> do
         let prefix = root -/- stageString stage -/- "lib"
             go gen file = generate file (semiEmptyTarget stage) gen
-        (prefix -/- "ghcplatform.h") %> go generateGhcPlatformH
         (prefix -/- "settings") %> go generateSettings
-        (prefix -/- "ghcautoconf.h") %> go generateGhcAutoconfH
-        (prefix -/- "ghcversion.h") %> go generateGhcVersionH
-        -- TODO: simplify, get rid of fake rts context
-        for_ (fst <$> deriveConstantsPairs) $ \constantsFile ->
-            prefix -/- constantsFile %> \file -> do
-                -- N.B. deriveConstants needs to compile programs which #include
-                -- PosixSource.h, which #include's ghcplatform.h. Fixes #18290.
-                need
-                    [ prefix -/- "ghcplatform.h"
-                    , prefix -/- "ghcautoconf.h"
-                    , prefix -/- "ghcversion.h"
-                    ]
-                withTempDir $ \dir -> build $
-                    target (rtsContext stage) DeriveConstants [] [file, dir]
+
   where
     file <~+ gen = file %> \out -> generate out emptyTarget gen >> makeExecutable out
 
@@ -212,19 +236,122 @@ emptyTarget :: Context
 emptyTarget = vanillaContext (error "Rules.Generate.emptyTarget: unknown stage")
                              (error "Rules.Generate.emptyTarget: unknown package")
 
+-- | A set of interpolation variable substitutions.
+newtype Interpolations = Interpolations (Action [(String, String)])
+
+instance Semigroup Interpolations where
+    Interpolations m <> Interpolations n = Interpolations ((++) <$> m <*> n)
+
+instance Monoid Interpolations where
+    mempty = Interpolations $ return []
+
+-- | @interpolateVar var value@ is an interpolation which replaces @\@var\@@
+-- with the result of @value@.
+interpolateVar :: String -> Action String -> Interpolations
+interpolateVar var value = Interpolations $ do
+    val <- value
+    return [(var, val)]
+
+runInterpolations :: Interpolations -> String -> Action String
+runInterpolations (Interpolations mk_substs) input = do
+    substs <- mk_substs
+    let subst :: String -> String
+        subst = foldr (.) id [replace ("@"++k++"@") v | (k,v) <- substs]
+    return (subst input)
+
+toCabalBool :: Bool -> String
+toCabalBool True  = "True"
+toCabalBool False = "False"
+
+-- | Interpolate the given variable with the value of the given 'Flag', using
+-- Cabal's boolean syntax.
+interpolateCabalFlag :: String -> Flag -> Interpolations
+interpolateCabalFlag name flg = interpolateVar name $ do
+    val <- flag flg
+    return (toCabalBool val)
+
+-- | Interpolate the given variable with the value of the given 'Setting'.
+interpolateSetting :: String -> Setting -> Interpolations
+interpolateSetting name settng = interpolateVar name $ setting settng
+
+-- | Interpolate the @ProjectVersion@ and @ProjectVersionMunged@ variables.
+projectVersion :: Interpolations
+projectVersion = mconcat
+    [ interpolateSetting "ProjectVersion" ProjectVersion
+    , interpolateSetting "ProjectVersionMunged" ProjectVersionMunged
+    ]
+
+rtsCabalFlags :: Interpolations
+rtsCabalFlags = mconcat
+    [ flag "CabalHaveLibdw" UseLibdw
+    , flag "CabalHaveLibm" UseLibm
+    , flag "CabalHaveLibrt" UseLibrt
+    , flag "CabalHaveLibdl" UseLibdl
+    , flag "CabalNeedLibpthread" UseLibpthread
+    , flag "CabalHaveLibbfd" UseLibbfd
+    , flag "CabalHaveLibNuma" UseLibnuma
+    , flag "CabalNeedLibatomic" NeedLibatomic
+    , flag "CabalUseSystemLibFFI" UseSystemFfi
+    , flag "CabalLibffiAdjustors" UseLibffiForAdjustors
+    , flag "CabalLeadingUnderscore" LeadingUnderscore
+    , interpolateVar "Cabal64bit" $ do
+        let settingWord :: Setting -> Action Word
+            settingWord s = read <$> setting s
+        ws <- settingWord TargetWordSize
+        return $ toCabalBool (ws == 8)
+    ]
+  where
+    flag = interpolateCabalFlag
+
+packageVersions :: Interpolations
+packageVersions = foldMap f [ base, ghcPrim, ghc, cabal, templateHaskell, ghcCompact, array ]
+  where
+    f :: Package -> Interpolations
+    f pkg = interpolateVar var $ show . version <$> readPackageData pkg
+      where var = "LIBRARY_" <> pkgName pkg <> "_VERSION"
+
+templateRule :: FilePath -> Interpolations -> Rules ()
+templateRule outPath interps = do
+    outPath %> \_ -> do
+        s <- readFile' (outPath <.> "in")
+        result <- runInterpolations interps s
+        writeFile' outPath result
+        putSuccess ("| Successfully generated " ++ outPath ++ " from its template")
+
+templateRules :: Rules ()
+templateRules = do
+  templateRule "compiler/ghc.cabal" $ projectVersion
+  templateRule "rts/rts.cabal" $ rtsCabalFlags
+  templateRule "driver/ghci/ghci-wrapper.cabal" $ projectVersion
+  templateRule "ghc/ghc-bin.cabal" $ projectVersion
+  templateRule "utils/iserv/iserv.cabal" $ projectVersion
+  templateRule "utils/iserv-proxy/iserv-proxy.cabal" $ projectVersion
+  templateRule "utils/remote-iserv/remote-iserv.cabal" $ projectVersion
+  templateRule "utils/runghc/runghc.cabal" $ projectVersion
+  templateRule "libraries/ghc-boot/ghc-boot.cabal" $ projectVersion
+  templateRule "libraries/ghc-boot-th/ghc-boot-th.cabal" $ projectVersion
+  templateRule "libraries/ghci/ghci.cabal" $ projectVersion
+  templateRule "libraries/ghc-heap/ghc-heap.cabal" $ projectVersion
+  templateRule "utils/ghc-pkg/ghc-pkg.cabal" $ projectVersion
+  templateRule "libraries/libiserv/libiserv.cabal" $ projectVersion
+  templateRule "libraries/template-haskell/template-haskell.cabal" $ projectVersion
+  templateRule "libraries/prologue.txt" $ packageVersions
+
+
 -- Generators
 
 -- | GHC wrapper scripts used for passing the path to the right package database
 -- when invoking in-tree GHC executables.
 ghcWrapper :: Stage -> Expr String
-ghcWrapper Stage0 = error "Stage0 GHC does not require a wrapper script to run."
+ghcWrapper (Stage0 {}) = error "Stage0 GHC does not require a wrapper script to run."
 ghcWrapper stage  = do
-    dbPath  <- expr $ (</>) <$> topDirectory <*> packageDbPath stage
+    dbPath  <- expr $ (</>) <$> topDirectory <*> packageDbPath (PackageDbLoc stage Final)
     ghcPath <- expr $ (</>) <$> topDirectory
-                            <*> programPath (vanillaContext (pred stage) ghc)
+                            <*> programPath (vanillaContext (predStage stage) ghc)
     return $ unwords $ map show $ [ ghcPath ]
                                ++ (if stage == Stage1
                                      then ["-no-global-package-db"
+                                          , "-package-env=-"
                                           , "-package-db " ++ dbPath
                                           ]
                                      else [])
@@ -240,7 +367,7 @@ generateGhcPlatformH :: Expr String
 generateGhcPlatformH = do
     trackGenerateHs
     stage    <- getStage
-    let chooseSetting x y = getSetting $ if stage == Stage0 then x else y
+    let chooseSetting x y = getSetting $ case stage of { Stage0 {} -> x; _ -> y }
     buildPlatform  <- chooseSetting BuildPlatform HostPlatform
     buildArch      <- chooseSetting BuildArch     HostArch
     buildOs        <- chooseSetting BuildOs       HostOs
@@ -250,13 +377,10 @@ generateGhcPlatformH = do
     hostOs         <- chooseSetting HostOs        TargetOs
     hostVendor     <- chooseSetting HostVendor    TargetVendor
     ghcUnreg       <- getFlag    GhcUnregisterised
-    inplaceTools   <- getFlag    SystemDistroMINGW
     return . unlines $
         [ "#if !defined(__GHCPLATFORM_H__)"
         , "#define __GHCPLATFORM_H__"
         , ""
-        , "#define GHC_STAGE " ++ show (fromEnum stage + 1)
-        , ""
         , "#define BuildPlatform_TYPE  " ++ cppify buildPlatform
         , "#define HostPlatform_TYPE   " ++ cppify hostPlatform
         , ""
@@ -280,8 +404,6 @@ generateGhcPlatformH = do
         , ""
         ]
         ++
-        [ "#define USE_INPLACE_MINGW_TOOLCHAIN 1" | inplaceTools ]
-        ++
         [ "#define UnregisterisedCompiler 1" | ghcUnreg ]
         ++
         [ ""
@@ -293,9 +415,10 @@ generateSettings :: Expr String
 generateSettings = do
     ctx <- getContext
     settings <- traverse sequence $
-        [ ("GCC extra via C opts", expr $ lookupValueOrError configFile "gcc-extra-via-c-opts")
+        [ ("GCC extra via C opts", expr $ lookupSystemConfig "gcc-extra-via-c-opts")
         , ("C compiler command", expr $ settingsFileSetting SettingsFileSetting_CCompilerCommand)
         , ("C compiler flags", expr $ settingsFileSetting SettingsFileSetting_CCompilerFlags)
+        , ("C++ compiler command", expr $ settingsFileSetting SettingsFileSetting_CxxCompilerCommand)
         , ("C++ compiler flags", expr $ settingsFileSetting SettingsFileSetting_CxxCompilerFlags)
         , ("C compiler link flags", expr $ settingsFileSetting SettingsFileSetting_CCompilerLinkFlags)
         , ("C compiler supports -no-pie", expr $ settingsFileSetting SettingsFileSetting_CCompilerSupportsNoPie)
@@ -303,46 +426,47 @@ generateSettings = do
         , ("Haskell CPP flags", expr $ settingsFileSetting SettingsFileSetting_HaskellCPPFlags)
         , ("ld command", expr $ settingsFileSetting SettingsFileSetting_LdCommand)
         , ("ld flags", expr $ settingsFileSetting SettingsFileSetting_LdFlags)
-        , ("ld supports compact unwind", expr $ lookupValueOrError configFile "ld-has-no-compact-unwind")
-        , ("ld supports build-id", expr $ lookupValueOrError configFile "ld-has-build-id")
-        , ("ld supports filelist", expr $ lookupValueOrError configFile "ld-has-filelist")
-        , ("ld is GNU ld", expr $ lookupValueOrError configFile "ld-is-gnu-ld")
+        , ("ld supports compact unwind", expr $ lookupSystemConfig "ld-has-no-compact-unwind")
+        , ("ld supports filelist", expr $ lookupSystemConfig "ld-has-filelist")
+        , ("ld is GNU ld", expr $ lookupSystemConfig "ld-is-gnu-ld")
         , ("Merge objects command", expr $ settingsFileSetting SettingsFileSetting_MergeObjectsCommand)
         , ("Merge objects flags", expr $ settingsFileSetting SettingsFileSetting_MergeObjectsFlags)
         , ("ar command", expr $ settingsFileSetting SettingsFileSetting_ArCommand)
-        , ("ar flags", expr $ lookupValueOrError configFile "ar-args")
+        , ("ar flags", expr $ lookupSystemConfig "ar-args")
         , ("ar supports at file", expr $ yesNo <$> flag ArSupportsAtFile)
+        , ("ar supports -L", expr $ yesNo <$> flag ArSupportsDashL)
         , ("ranlib command", expr $ settingsFileSetting SettingsFileSetting_RanlibCommand)
         , ("otool command", expr $ settingsFileSetting SettingsFileSetting_OtoolCommand)
         , ("install_name_tool command", expr $ settingsFileSetting SettingsFileSetting_InstallNameToolCommand)
         , ("touch command", expr $ settingsFileSetting SettingsFileSetting_TouchCommand)
         , ("dllwrap command", expr $ settingsFileSetting SettingsFileSetting_DllWrapCommand)
         , ("windres command", expr $ settingsFileSetting SettingsFileSetting_WindresCommand)
-        , ("libtool command", expr $ settingsFileSetting SettingsFileSetting_LibtoolCommand)
         , ("unlit command", ("$topdir/bin/" <>) <$> expr (programName (ctx { Context.package = unlit })))
         , ("cross compiling", expr $ yesNo <$> flag CrossCompiling)
         , ("target platform string", getSetting TargetPlatform)
         , ("target os", getSetting TargetOsHaskell)
         , ("target arch", getSetting TargetArchHaskell)
-        , ("target word size", expr $ lookupValueOrError configFile "target-word-size")
-        , ("target word big endian", expr $ lookupValueOrError configFile "target-word-big-endian")
-        , ("target has GNU nonexec stack", expr $ lookupValueOrError configFile "target-has-gnu-nonexec-stack")
-        , ("target has .ident directive", expr $ lookupValueOrError configFile "target-has-ident-directive")
-        , ("target has subsections via symbols", expr $ lookupValueOrError configFile "target-has-subsections-via-symbols")
-        , ("target has RTS linker", expr $ lookupValueOrError configFile "target-has-rts-linker")
+        , ("target word size", expr $ lookupSystemConfig "target-word-size")
+        , ("target word big endian", expr $ lookupSystemConfig "target-word-big-endian")
+        , ("target has GNU nonexec stack", expr $ lookupSystemConfig "target-has-gnu-nonexec-stack")
+        , ("target has .ident directive", expr $ lookupSystemConfig "target-has-ident-directive")
+        , ("target has subsections via symbols", expr $ lookupSystemConfig "target-has-subsections-via-symbols")
+        , ("target has RTS linker", expr $ lookupSystemConfig "target-has-rts-linker")
+        , ("target has libm", expr $  lookupSystemConfig "target-has-libm")
         , ("Unregisterised", expr $ yesNo <$> flag GhcUnregisterised)
         , ("LLVM target", getSetting LlvmTarget)
         , ("LLVM llc command", expr $ settingsFileSetting SettingsFileSetting_LlcCommand)
         , ("LLVM opt command", expr $ settingsFileSetting SettingsFileSetting_OptCommand)
         , ("LLVM clang command", expr $ settingsFileSetting SettingsFileSetting_ClangCommand)
+        , ("Use inplace MinGW toolchain", expr $ settingsFileSetting SettingsFileSetting_DistroMinGW)
 
         , ("Use interpreter", expr $ yesNo <$> ghcWithInterpreter)
         , ("Support SMP", expr $ yesNo <$> targetSupportsSMP)
-        , ("RTS ways", unwords . map show <$> getRtsWays)
+        , ("RTS ways", unwords . map show . Set.toList <$> getRtsWays)
         , ("Tables next to code", expr $ yesNo <$> flag TablesNextToCode)
         , ("Leading underscore", expr $ yesNo <$> flag LeadingUnderscore)
         , ("Use LibFFI", expr $ yesNo <$> useLibffiForAdjustors)
-        , ("RTS expects libdw", yesNo <$> getFlag WithLibdw)
+        , ("RTS expects libdw", yesNo <$> getFlag UseLibdw)
         ]
     let showTuple (k, v) = "(" ++ show k ++ ", " ++ show v ++ ")"
     pure $ case settings of
@@ -357,15 +481,14 @@ generateSettings = do
 generateConfigHs :: Expr String
 generateConfigHs = do
     stage <- getStage
-    let chooseSetting x y = getSetting $ if stage == Stage0 then x else y
+    let chooseSetting x y = getSetting $ case stage of { Stage0 {} -> x; _ -> y }
     buildPlatform <- chooseSetting BuildPlatform HostPlatform
     hostPlatform <- chooseSetting HostPlatform TargetPlatform
     trackGenerateHs
     cProjectName        <- getSetting ProjectName
     cBooterVersion      <- getSetting GhcVersion
     return $ unlines
-        [ "{-# LANGUAGE CPP #-}"
-        , "module GHC.Settings.Config"
+        [ "module GHC.Settings.Config"
         , "  ( module GHC.Version"
         , "  , cBuildPlatformString"
         , "  , cHostPlatformString"
@@ -374,7 +497,7 @@ generateConfigHs = do
         , "  , cStage"
         , "  ) where"
         , ""
-        , "import GHC.Prelude"
+        , "import GHC.Prelude.Basic"
         , ""
         , "import GHC.Version"
         , ""
@@ -391,14 +514,21 @@ generateConfigHs = do
         , "cBooterVersion        = " ++ show cBooterVersion
         , ""
         , "cStage                :: String"
-        , "cStage                = show (" ++ show (fromEnum stage + 1) ++ " :: Int)"
+        , "cStage                = show (" ++ stageString stage ++ " :: Int)"
         ]
+  where
+    stageString (Stage0 InTreeLibs) = "1"
+    stageString Stage1 = "2"
+    stageString Stage2 = "3"
+    stageString Stage3 = "4"
+    stageString (Stage0 GlobalLibs) = error "stageString: StageBoot"
+
 
 -- | Generate @ghcautoconf.h@ header.
 generateGhcAutoconfH :: Expr String
 generateGhcAutoconfH = do
     trackGenerateHs
-    configHContents  <- expr $ map undefinePackage <$> readFileLines configH
+    configHContents  <- expr $ mapMaybe undefinePackage <$> readFileLines configH
     return . unlines $
         [ "#if !defined(__GHCAUTOCONF_H__)"
         , "#define __GHCAUTOCONF_H__" ]
@@ -407,43 +537,12 @@ generateGhcAutoconfH = do
   where
     undefinePackage s
         | "#define PACKAGE_" `isPrefixOf` s
-            = "/* #undef " ++ takeWhile (/=' ') (drop 8 s) ++ " */"
-        | otherwise = s
-
--- | Generate @ghcversion.h@ header.
-generateGhcVersionH :: Expr String
-generateGhcVersionH = do
-    trackGenerateHs
-    fullVersion <- getSetting ProjectVersion
-    version     <- getSetting ProjectVersionInt
-    patchLevel1 <- getSetting ProjectPatchLevel1
-    patchLevel2 <- getSetting ProjectPatchLevel2
-    return . unlines $
-        [ "#if !defined(__GHCVERSION_H__)"
-        , "#define __GHCVERSION_H__"
-        , ""
-        , "#if !defined(__GLASGOW_HASKELL__)"
-        , "#define __GLASGOW_HASKELL__ " ++ version
-        , "#endif"
-        , "#if !defined(__GLASGOW_HASKELL_FULL_VERSION__)"
-        , "#define __GLASGOW_HASKELL_FULL_VERSION__ \"" ++ fullVersion ++ "\""
-        , "#endif"
-        , ""]
-        ++
-        [ "#define __GLASGOW_HASKELL_PATCHLEVEL1__ " ++ patchLevel1 | patchLevel1 /= "" ]
-        ++
-        [ "#define __GLASGOW_HASKELL_PATCHLEVEL2__ " ++ patchLevel2 | patchLevel2 /= "" ]
-        ++
-        [ ""
-        , "#define MIN_VERSION_GLASGOW_HASKELL(ma,mi,pl1,pl2) (\\"
-        , "   ((ma)*100+(mi)) <  __GLASGOW_HASKELL__ || \\"
-        , "   ((ma)*100+(mi)) == __GLASGOW_HASKELL__    \\"
-        , "          && (pl1) <  __GLASGOW_HASKELL_PATCHLEVEL1__ || \\"
-        , "   ((ma)*100+(mi)) == __GLASGOW_HASKELL__    \\"
-        , "          && (pl1) == __GLASGOW_HASKELL_PATCHLEVEL1__ \\"
-        , "          && (pl2) <= __GLASGOW_HASKELL_PATCHLEVEL2__ )"
-        , ""
-        , "#endif /* __GHCVERSION_H__ */" ]
+            = Just $ "/* #undef " ++ takeWhile (/=' ') (drop 8 s) ++ " */"
+        | "#define __GLASGOW_HASKELL" `isPrefixOf` s
+            = Nothing
+        | "/* REMOVE ME */" == s
+            = Nothing
+        | otherwise = Just s
 
 -- | Generate @Version.hs@ files.
 generateVersionHs :: Expr String
