@@ -423,6 +423,15 @@ printClosure( const StgClosure *obj )
         debugBelch("TREC_CHUNK\n");
         break;
 
+    case CONTINUATION:
+    {
+        StgContinuation *u = (StgContinuation *)obj;
+        debugBelch("CONTINUATION(apply_mask_frame=");
+        printPtr((StgPtr)u->apply_mask_frame);
+        debugBelch(",stack_size=%" FMT_Word ")\n", u->stack_size);
+        break;
+    }
+
     default:
             //barf("printClosure %d",get_itbl(obj)->type);
             debugBelch("*** printClosure: unknown type %d ****\n",
@@ -563,6 +572,8 @@ printStackChunk( StgPtr sp, StgPtr spBottom )
                 debugBelch("stg_ret_d_info\n" );
             } else if (c == (StgWord)&stg_ret_l_info) {
                 debugBelch("stg_ret_l_info\n" );
+            } else if (c == (StgWord)&stg_prompt_frame_info) {
+                debugBelch("stg_prompt_frame_info\n");
 #if defined(PROFILING)
             } else if (c == (StgWord)&stg_restore_cccs_info) {
                 debugBelch("stg_restore_cccs_info\n" );
@@ -703,9 +714,9 @@ void printWeakLists()
 {
     debugBelch("======= WEAK LISTS =======\n");
 
-    for (uint32_t cap_idx = 0; cap_idx < n_capabilities; ++cap_idx) {
+    for (uint32_t cap_idx = 0; cap_idx < getNumCapabilities(); ++cap_idx) {
         debugBelch("Capability %d:\n", cap_idx);
-        Capability *cap = capabilities[cap_idx];
+        Capability *cap = getCapability(cap_idx);
         for (StgWeak *weak = cap->weak_ptr_list_hd; weak; weak = weak->link) {
             printClosure((StgClosure*)weak);
         }
@@ -730,8 +741,8 @@ void printLargeAndPinnedObjects()
 {
     debugBelch("====== PINNED OBJECTS ======\n");
 
-    for (uint32_t cap_idx = 0; cap_idx < n_capabilities; ++cap_idx) {
-        Capability *cap = capabilities[cap_idx];
+    for (uint32_t cap_idx = 0; cap_idx < getNumCapabilities(); ++cap_idx) {
+        Capability *cap = getCapability(cap_idx);
 
         debugBelch("Capability %d: Current pinned object block: %p\n",
                    cap_idx, (void*)cap->pinned_object_block);
@@ -943,7 +954,7 @@ findPtr(P_ p, int follow)
   // We can't search the nursery, because we don't know which blocks contain
   // valid data, because the bd->free pointers in the nursery are only reset
   // just before a block is used.
-  for (n = 0; n < n_capabilities; n++) {
+  for (n = 0; n < getNumCapabilities(); n++) {
       bd = nurseries[i].blocks;
       i = findPtrBlocks(p,bd,arr,arr_size,i);
       if (i >= arr_size) return;
@@ -956,7 +967,7 @@ findPtr(P_ p, int follow)
       bd = generations[g].large_objects;
       i = findPtrBlocks(p,bd,arr,arr_size,i);
       if (i >= arr_size) return;
-      for (n = 0; n < n_capabilities; n++) {
+      for (n = 0; n < getNumCapabilities(); n++) {
           i = findPtrBlocks(p, gc_threads[n]->gens[g].part_list,
                             arr, arr_size, i);
           i = findPtrBlocks(p, gc_threads[n]->gens[g].todo_bd,
@@ -1063,10 +1074,11 @@ const char *closure_type_names[] = {
  [SMALL_MUT_ARR_PTRS_DIRTY] = "SMALL_MUT_ARR_PTRS_DIRTY",
  [SMALL_MUT_ARR_PTRS_FROZEN_DIRTY] = "SMALL_MUT_ARR_PTRS_FROZEN_DIRTY",
  [SMALL_MUT_ARR_PTRS_FROZEN_CLEAN] = "SMALL_MUT_ARR_PTRS_FROZEN_CLEAN",
- [COMPACT_NFDATA]        = "COMPACT_NFDATA"
+ [COMPACT_NFDATA]        = "COMPACT_NFDATA",
+ [CONTINUATION]          = "CONTINUATION",
 };
 
-#if N_CLOSURE_TYPES != 64
+#if N_CLOSURE_TYPES != 65
 #error Closure types changed: update Printer.c!
 #endif
 
