@@ -22,10 +22,6 @@
 
 /* ----------------------------------------------------------------------------
    Building Haskell objects from C datatypes.
-
-   TODO: Currently this code does not tag created pointers,
-         however it is not unsafe (the constructor code will do it)
-         just inefficient.
    ------------------------------------------------------------------------- */
 HaskellObj
 rts_mkChar (Capability *cap, HsChar c)
@@ -866,7 +862,7 @@ void rts_listMiscRoots (ListRootsCb cb, void *user)
 }
 
 #else
-PauseToken GNU_ATTRIBUTE(__noreturn__)
+PauseToken STG_NORETURN
 *rts_pause (void)
 {
     errorBelch("Warning: Pausing the RTS is only possible for "
@@ -874,7 +870,7 @@ PauseToken GNU_ATTRIBUTE(__noreturn__)
     stg_exit(EXIT_FAILURE);
 }
 
-void GNU_ATTRIBUTE(__noreturn__)
+void STG_NORETURN
 rts_resume (PauseToken *pauseToken STG_UNUSED)
 {
     errorBelch("Warning: Resuming the RTS is only possible for "
@@ -882,7 +878,7 @@ rts_resume (PauseToken *pauseToken STG_UNUSED)
     stg_exit(EXIT_FAILURE);
 }
 
-bool rts_isPaused(void)
+bool rts_isPaused()
 {
     errorBelch("Warning: Pausing/Resuming the RTS is only possible for "
                "multithreaded RTS.");
