@@ -4,6 +4,9 @@
  *
  * Generational garbage collector: evacuation functions
  *
+ * Evacuating is copying (live) objects from a source generation to the target
+ * generation.
+ *
  * Documentation on the architecture of the Garbage Collector can be
  * found in the online commentary:
  *
@@ -636,7 +639,8 @@ evacuate_compact (StgPtr p)
 /* ----------------------------------------------------------------------------
    Evacuate
 
-   This is called (eventually) for every live object in the system.
+   Copies an object from its current generation to a target generation. This is
+   called (eventually) for every live object in the system.
 
    The caller to evacuate specifies a desired generation in the
    gct->evac_gen thread-local variable.  The following conditions apply to
@@ -1058,6 +1062,10 @@ loop:
       copy(p,info,q,sizeofW(StgTRecChunk),gen_no);
       return;
 
+  case CONTINUATION:
+      copy(p,info,q,continuation_sizeW((StgContinuation*)q),gen_no);
+      return;
+
   default:
     barf("evacuate: strange closure type %d", (int)(INFO_PTR_TO_STRUCT(info)->type));
   }
