@@ -292,6 +292,7 @@ GarbageCollect (uint32_t collect_gen,
       any_work, scav_find_work, max_n_todo_overflow;
 #if defined(THREADED_RTS)
   gc_thread *saved_gct;
+  bool gc_sparks_all_caps;
 #endif
   uint32_t g, n;
   // The time we should report our heap census as occurring at, if necessary.
@@ -309,7 +310,7 @@ GarbageCollect (uint32_t collect_gen,
 #endif
 
 #if defined(PROFILING)
-  CostCentreStack *save_CCS[n_capabilities];
+  CostCentreStack *save_CCS[getNumCapabilities()];
 #endif
 
   ACQUIRE_SM_LOCK;
@@ -340,9 +341,9 @@ GarbageCollect (uint32_t collect_gen,
 
   // attribute any costs to CCS_GC
 #if defined(PROFILING)
-  for (n = 0; n < n_capabilities; n++) {
-      save_CCS[n] = capabilities[n]->r.rCCCS;
-      capabilities[n]->r.rCCCS = CCS_GC;
+  for (n = 0; n < getNumCapabilities(); n++) {
+      save_CCS[n] = getCapability(n)->r.rCCCS;
+      getCapability(n)->r.rCCCS = CCS_GC;
   }
 #endif
 
@@ -355,7 +356,7 @@ GarbageCollect (uint32_t collect_gen,
   deadlock_detect_gc = deadlock_detect;
 
 #if defined(THREADED_RTS)
-  if (major_gc && RtsFlags.GcFlags.useNonmoving && concurrent_coll_running) {
+  if (major_gc && RtsFlags.GcFlags.useNonmoving && RELAXED_LOAD(&concurrent_coll_running)) {
       /* If there is already a concurrent major collection running then
        * there is no benefit to starting another.
        * TODO: Catch heap-size runaway.
@@ -396,9 +397,9 @@ GarbageCollect (uint32_t collect_gen,
    * here
   */
   if (gc_type == SYNC_GC_PAR) {
-      n_gc_threads = n_capabilities;
+      n_gc_threads = getNumCapabilities();
       n_gc_idle_threads = 0;
-      for (uint32_t i = 0; i < n_capabilities; ++i) {
+      for (uint32_t i = 0; i < getNumCapabilities(); ++i) {
           if (idle_cap[i]) {
               ASSERT(i != gct->thread_index);
               ++n_gc_idle_threads;
@@ -406,7 +407,7 @@ GarbageCollect (uint32_t collect_gen,
       }
   } else {
       n_gc_threads = 1;
-      n_gc_idle_threads = n_capabilities - 1;
+      n_gc_idle_threads = getNumCapabilities() - 1;
   }
   work_stealing = RtsFlags.ParFlags.parGcLoadBalancingEnabled &&
       N >= RtsFlags.ParFlags.parGcLoadBalancingGen &&
@@ -429,15 +430,15 @@ GarbageCollect (uint32_t collect_gen,
   SEQ_CST_STORE(&gc_running_threads, 0);
 
   ASSERT(n_gc_threads > 0);
-  ASSERT(n_gc_threads <= n_capabilities);
-  ASSERT(n_gc_idle_threads < n_capabilities);
+  ASSERT(n_gc_threads <= getNumCapabilities());
+  ASSERT(n_gc_idle_threads < getNumCapabilities());
   // If we are work stealing, there better be another(i.e. not us) non-idle gc
   // thread
   ASSERT(!work_stealing || n_gc_threads - 1 > n_gc_idle_threads);
 
 
   debugTrace(DEBUG_gc, "GC (gen %d, using %d thread(s), %s work stealing)",
-             N, (int)n_capabilities - (int)n_gc_idle_threads,
+             N, (int)getNumCapabilities() - (int)n_gc_idle_threads,
              work_stealing ? "with": "without");
 
 #if defined(DEBUG)
@@ -445,6 +446,11 @@ GarbageCollect (uint32_t collect_gen,
   memInventory(DEBUG_gc);
 #endif
 
+  // Defer all free calls for the megablock allocator to avoid quadratic runtime
+  // explosion when freeing a lot of memory in a single GC
+  // (https://gitlab.haskell.org/ghc/ghc/-/issues/19897).
+  deferMBlockFreeing();
+
   // do this *before* we start scavenging
   collectFreshWeakPtrs();
 
@@ -499,20 +505,20 @@ GarbageCollect (uint32_t collect_gen,
   // call back into the GC via mark_root() (due to the gct register
   // variable).
   if (!is_par_gc()) {
-      for (n = 0; n < n_capabilities; n++) {
+      for (n = 0; n < getNumCapabilities(); n++) {
 #if defined(THREADED_RTS)
-          scavenge_capability_mut_Lists1(capabilities[n]);
+          scavenge_capability_mut_Lists1(getCapability(n));
 #else
-          scavenge_capability_mut_lists(capabilities[n]);
+          scavenge_capability_mut_lists(getCapability(n));
 #endif
       }
   } else {
       scavenge_capability_mut_lists(gct->cap);
-      for (n = 0; n < n_capabilities; n++) {
+      for (n = 0; n < getNumCapabilities(); n++) {
           if (idle_cap[n]) {
-              markCapability(mark_root, gct, capabilities[n],
+              markCapability(mark_root, gct, getCapability(n),
                              true/*don't mark sparks*/);
-              scavenge_capability_mut_lists(capabilities[n]);
+              scavenge_capability_mut_lists(getCapability(n));
           }
       }
   }
@@ -524,16 +530,14 @@ GarbageCollect (uint32_t collect_gen,
   // follow all the roots that the application knows about.
   gct->evac_gen_no = 0;
   if (!is_par_gc()) {
-      for (n = 0; n < n_capabilities; n++) {
-          markCapability(mark_root, gct, capabilities[n],
+      for (n = 0; n < getNumCapabilities(); n++) {
+          markCapability(mark_root, gct, getCapability(n),
                          true/*don't mark sparks*/);
       }
   } else {
       markCapability(mark_root, gct, cap, true/*don't mark sparks*/);
   }
 
-  markScheduler(mark_root, gct);
-
   // Mark the weak pointer list, and prepare to detect dead weak pointers.
   markWeakPtrList();
   initWeakForGC();
@@ -556,6 +560,9 @@ GarbageCollect (uint32_t collect_gen,
   StgTSO *resurrected_threads = END_TSO_QUEUE;
   // must be last...  invariant is that everything is fully
   // scavenged at this point.
+#if defined(THREADED_RTS)
+  gc_sparks_all_caps = !work_stealing || !is_par_gc();
+#endif
   work_stealing = false;
   while (traverseWeakPtrList(&dead_weak_ptr_list, &resurrected_threads))
   {
@@ -568,14 +575,15 @@ GarbageCollect (uint32_t collect_gen,
   gcStableNameTable();
 
 #if defined(THREADED_RTS)
-  if (!is_par_gc()) {
+  // See Note [Pruning the spark pool]
+  if(gc_sparks_all_caps) {
       for (n = 0; n < n_capabilities; n++) {
-          pruneSparkQueue(false, capabilities[n]);
+          pruneSparkQueue(false, getCapability(n));
       }
   } else {
-      for (n = 0; n < n_capabilities; n++) {
+      for (n = 0; n < getNumCapabilities(); n++) {
           if (n == cap->no || idle_cap[n]) {
-              pruneSparkQueue(false, capabilities[n]);
+              pruneSparkQueue(false, getCapability(n));
          }
       }
   }
@@ -679,8 +687,8 @@ GarbageCollect (uint32_t collect_gen,
     // stats.  Every mutable list is copied during every GC.
     if (g > 0) {
         W_ mut_list_size = 0;
-        for (n = 0; n < n_capabilities; n++) {
-            mut_list_size += countOccupied(capabilities[n]->mut_lists[g]);
+        for (n = 0; n < getNumCapabilities(); n++) {
+            mut_list_size += countOccupied(getCapability(n)->mut_lists[g]);
         }
         copied +=  mut_list_size;
 
@@ -830,23 +838,18 @@ GarbageCollect (uint32_t collect_gen,
     live_blocks += genLiveBlocks(gen);
 
     // add in the partial blocks in the gen_workspaces
-    {
-        uint32_t i;
-        for (i = 0; i < n_capabilities; i++) {
-            live_words  += gcThreadLiveWords(i, gen->no);
-            live_blocks += gcThreadLiveBlocks(i, gen->no);
-        }
+    for (uint32_t i = 0; i < getNumCapabilities(); i++) {
+        live_words  += gcThreadLiveWords(i, gen->no);
+        live_blocks += gcThreadLiveBlocks(i, gen->no);
     }
   } // for all generations
 
   // Flush the update remembered sets. See Note [Eager update remembered set
   // flushing] in NonMovingMark.c
   if (RtsFlags.GcFlags.useNonmoving) {
-      RELEASE_SM_LOCK;
-      for (n = 0; n < n_capabilities; n++) {
-          nonmovingAddUpdRemSetBlocks(&capabilities[n]->upd_rem_set.queue);
+      for (n = 0; n < getNumCapabilities(); n++) {
+          nonmovingAddUpdRemSetBlocks(&getCapability(n)->upd_rem_set);
       }
-      ACQUIRE_SM_LOCK;
   }
 
   // Mark and sweep the oldest generation.
@@ -867,8 +870,6 @@ GarbageCollect (uint32_t collect_gen,
       // old_weak_ptr_list should be empty.
       ASSERT(oldest_gen->old_weak_ptr_list == NULL);
 
-      // we may need to take the lock to allocate mark queue blocks
-      RELEASE_SM_LOCK;
       // dead_weak_ptr_list contains weak pointers with dead keys. Those need to
       // be kept alive because we'll use them in finalizeSchedulers(). Similarly
       // resurrected_threads are also going to be used in resurrectedThreads()
@@ -878,10 +879,9 @@ GarbageCollect (uint32_t collect_gen,
 #if !defined(THREADED_RTS)
       // In the non-threaded runtime this is the only time we push to the
       // upd_rem_set
-      nonmovingAddUpdRemSetBlocks(&gct->cap->upd_rem_set.queue);
+      nonmovingAddUpdRemSetBlocks(&gct->cap->upd_rem_set);
 #endif
       nonmovingCollect(&dead_weak_ptr_list, &resurrected_threads);
-      ACQUIRE_SM_LOCK;
   }
 
   // Update the max size of older generations after a major GC:
@@ -984,6 +984,11 @@ GarbageCollect (uint32_t collect_gen,
   resurrectThreads(resurrected_threads);
   ACQUIRE_SM_LOCK;
 
+  // Finally free the deferred mblocks by sorting the deferred free list and
+  // merging it into the actual sorted free list. This needs to happen here so
+  // that the `returnMemoryToOS` call down below can successfully free memory.
+  commitMBlockFreeing();
+
   if (major_gc) {
       W_ need_prealloc, need_live, need, got;
       uint32_t i;
@@ -1051,6 +1056,13 @@ GarbageCollect (uint32_t collect_gen,
           returned = returnMemoryToOS(got - need);
       }
       traceEventMemReturn(cap, got, need, returned);
+
+      // Ensure that we've returned enough mblocks to place us under maxHeapSize.
+      // This may fail for instance due to block fragmentation.
+      W_ after = got - returned;
+      if (RtsFlags.GcFlags.maxHeapSize != 0 && after > BLOCKS_TO_MBLOCKS(RtsFlags.GcFlags.maxHeapSize)) {
+        heapOverflow();
+      }
   }
 
   // extra GC trace info
@@ -1063,8 +1075,8 @@ GarbageCollect (uint32_t collect_gen,
 
   // restore enclosing cost centre
 #if defined(PROFILING)
-  for (n = 0; n < n_capabilities; n++) {
-      capabilities[n]->r.rCCCS = save_CCS[n];
+  for (n = 0; n < getNumCapabilities(); n++) {
+      getCapability(n)->r.rCCCS = save_CCS[n];
   }
 #endif
 
@@ -1113,7 +1125,7 @@ new_gc_thread (uint32_t n, gc_thread *t)
     uint32_t g;
     gen_workspace *ws;
 
-    t->cap = capabilities[n];
+    t->cap = getCapability(n);
 
 #if defined(THREADED_RTS)
     t->id = 0;
@@ -1212,7 +1224,7 @@ freeGcThreads (void)
     if (gc_threads != NULL) {
 #if defined(THREADED_RTS)
         uint32_t i;
-        for (i = 0; i < n_capabilities; i++) {
+        for (i = 0; i < getNumCapabilities(); i++) {
             for (g = 0; g < RtsFlags.GcFlags.generations; g++)
             {
                 freeWSDeque(gc_threads[i]->gens[g].todo_q);
@@ -1372,7 +1384,6 @@ void
 gcWorkerThread (Capability *cap)
 {
     gc_thread *saved_gct;
-
     // necessary if we stole a callee-saves register for gct:
     saved_gct = gct;
 
@@ -1403,13 +1414,10 @@ gcWorkerThread (Capability *cap)
     scavenge_until_all_done();
 
 #if defined(THREADED_RTS)
-    // Now that the whole heap is marked, we discard any sparks that
-    // were found to be unreachable.  The main GC thread is currently
-    // marking heap reachable via weak pointers, so it is
-    // non-deterministic whether a spark will be retained if it is
-    // only reachable via weak pointers.  To fix this problem would
-    // require another GC barrier, which is too high a price.
-    pruneSparkQueue(false, cap);
+    // See Note [Pruning the spark pool]
+    if(work_stealing && is_par_gc()) {
+        pruneSparkQueue(false, cap);
+    }
 #endif
 
     // Wait until we're told to continue
@@ -1442,31 +1450,31 @@ void
 waitForGcThreads (Capability *cap, bool idle_cap[])
 {
     // n_gc_threads is not valid here, we're too early
-    uint32_t n_threads = n_capabilities;
+    uint32_t n_threads = getNumCapabilities();
     const uint32_t me = cap->no;
     uint32_t i, cur_n_gc_entered;
     Time t0, t1, t2;
 
     t0 = t1 = t2 = getProcessElapsedTime();
 
-    for(i = 0; i < n_capabilities; ++i) {
+    for(i = 0; i < getNumCapabilities(); ++i) {
         if (i == me || idle_cap[i]) {
             --n_threads;
         }
     }
 
-    ASSERT(n_threads < n_capabilities); // must be less because we don't count ourself
+    ASSERT(n_threads < getNumCapabilities()); // must be less because we don't count ourself
     if(n_threads == 0) { return; }
 
     ACQUIRE_LOCK(&gc_entry_mutex);
     while((cur_n_gc_entered = SEQ_CST_LOAD(&n_gc_entered)) != n_threads) {
         ASSERT(cur_n_gc_entered < n_threads);
-        for(i = 0; i < n_capabilities; ++i) {
+        for(i = 0; i < getNumCapabilities(); ++i) {
             if (i == me || idle_cap[i]) { continue; }
             if (SEQ_CST_LOAD(&gc_threads[i]->wakeup) != GC_THREAD_STANDING_BY) {
-                prodCapability(capabilities[i], cap->running_task);
+                prodCapability(getCapability(i), cap->running_task);
                 write_barrier();
-                interruptCapability(capabilities[i]);
+                interruptCapability(getCapability(i));
             }
         }
         // this 1ms timeout is not well justified. It's the shortest timeout we
@@ -1551,7 +1559,7 @@ shutdown_gc_threads (uint32_t me USED_IF_THREADS USED_IF_DEBUG,
     }
 #if defined(DEBUG)
     uint32_t i;
-    for (i=0; i < n_capabilities; i++) {
+    for (i=0; i < getNumCapabilities(); i++) {
         if (i == me || idle_cap[i]) continue;
         ASSERT(SEQ_CST_LOAD(&gc_threads[i]->wakeup) == GC_THREAD_WAITING_TO_CONTINUE);
     }
@@ -1564,7 +1572,7 @@ shutdown_gc_threads (uint32_t me USED_IF_THREADS USED_IF_DEBUG,
 void
 releaseGCThreads (Capability *cap USED_IF_THREADS, bool idle_cap[])
 {
-    const uint32_t n_threads = n_capabilities;
+    const uint32_t n_threads = getNumCapabilities();
     const uint32_t me = cap->no;
     uint32_t i;
     uint32_t num_idle = 0;
@@ -1616,19 +1624,19 @@ prepare_collected_gen (generation *gen)
 
     if (RtsFlags.GcFlags.useNonmoving && g == oldest_gen->no) {
         // Nonmoving heap's mutable list is always a root.
-        for (i = 0; i < n_capabilities; i++) {
-            stash_mut_list(capabilities[i], g);
+        for (i = 0; i < getNumCapabilities(); i++) {
+            stash_mut_list(getCapability(i), g);
         }
     } else if (g != 0) {
         // Otherwise throw away the current mutable list. Invariant: the
         // mutable list always has at least one block; this means we can avoid
         // a check for NULL in recordMutable().
-        for (i = 0; i < n_capabilities; i++) {
-            bdescr *old = RELAXED_LOAD(&capabilities[i]->mut_lists[g]);
+        for (i = 0; i < getNumCapabilities(); i++) {
+            bdescr *old = RELAXED_LOAD(&getCapability(i)->mut_lists[g]);
             freeChain(old);
 
             bdescr *new = allocBlockOnNode(capNoToNumaNode(i));
-            RELAXED_STORE(&capabilities[i]->mut_lists[g], new);
+            RELAXED_STORE(&getCapability(i)->mut_lists[g], new);
         }
     }
 
@@ -1660,7 +1668,7 @@ prepare_collected_gen (generation *gen)
 
     // grab all the partial blocks stashed in the gc_thread workspaces and
     // move them to the old_blocks list of this gen.
-    for (n = 0; n < n_capabilities; n++) {
+    for (n = 0; n < getNumCapabilities(); n++) {
         ws = &gc_threads[n]->gens[gen->no];
 
         for (bd = ws->part_list; bd != NULL; bd = next) {
@@ -1759,8 +1767,8 @@ prepare_uncollected_gen (generation *gen)
     // save the current mutable lists for this generation, and
     // allocate a fresh block for each one.  We'll traverse these
     // mutable lists as roots early on in the GC.
-    for (i = 0; i < n_capabilities; i++) {
-        stash_mut_list(capabilities[i], gen->no);
+    for (i = 0; i < getNumCapabilities(); i++) {
+        stash_mut_list(getCapability(i), gen->no);
     }
 
     ASSERT(gen->scavenged_large_objects == NULL);
@@ -1833,11 +1841,11 @@ collect_pinned_object_blocks (void)
     const bool use_nonmoving = RtsFlags.GcFlags.useNonmoving;
     generation *const gen = (use_nonmoving && major_gc) ? oldest_gen : g0;
 
-    for (uint32_t n = 0; n < n_capabilities; n++) {
+    for (uint32_t n = 0; n < getNumCapabilities(); n++) {
         bdescr *last = NULL;
         if (use_nonmoving && gen == oldest_gen) {
             // Mark objects as belonging to the nonmoving heap
-            for (bdescr *bd = RELAXED_LOAD(&capabilities[n]->pinned_object_blocks); bd != NULL; bd = bd->link) {
+            for (bdescr *bd = RELAXED_LOAD(&getCapability(n)->pinned_object_blocks); bd != NULL; bd = bd->link) {
                 bd->flags |= BF_NONMOVING;
                 bd->gen = oldest_gen;
                 bd->gen_no = oldest_gen->no;
@@ -1846,7 +1854,7 @@ collect_pinned_object_blocks (void)
                 last = bd;
             }
         } else {
-            for (bdescr *bd = capabilities[n]->pinned_object_blocks; bd != NULL; bd = bd->link) {
+            for (bdescr *bd = getCapability(n)->pinned_object_blocks; bd != NULL; bd = bd->link) {
                 last = bd;
             }
         }
@@ -1856,8 +1864,8 @@ collect_pinned_object_blocks (void)
             if (gen->large_objects != NULL) {
                 gen->large_objects->u.back = last;
             }
-            gen->large_objects = RELAXED_LOAD(&capabilities[n]->pinned_object_blocks);
-            RELAXED_STORE(&capabilities[n]->pinned_object_blocks, NULL);
+            gen->large_objects = RELAXED_LOAD(&getCapability(n)->pinned_object_blocks);
+            RELAXED_STORE(&getCapability(n)->pinned_object_blocks, NULL);
         }
     }
 }
@@ -1949,7 +1957,7 @@ resizeGenerations (void)
     // minimum size for generation zero
     min_alloc = stg_max((RtsFlags.GcFlags.pcFreeHeap * max) / 200,
                         RtsFlags.GcFlags.minAllocAreaSize
-                        * (W_)n_capabilities);
+                        * (W_)getNumCapabilities());
 
     // Auto-enable compaction when the residency reaches a
     // certain percentage of the maximum heap size (default: 30%).
@@ -2021,7 +2029,7 @@ static void
 resize_nursery (void)
 {
     const StgWord min_nursery =
-      RtsFlags.GcFlags.minAllocAreaSize * (StgWord)n_capabilities;
+      RtsFlags.GcFlags.minAllocAreaSize * (StgWord)getNumCapabilities();
 
     if (RtsFlags.GcFlags.generations == 1)
     {   // Two-space collector:
@@ -2266,7 +2274,7 @@ bool doIdleGCWork(Capability *cap STG_UNUSED, bool all)
  * usage when the live data size is much more reasonable (for example ghcide)
  *
  * Therefore we have a new (2021) strategy which starts by retaining up to 4 * live_bytes
- * of blocks before gradually returning uneeded memory back to the OS on subsequent
+ * of blocks before gradually returning unneeded memory back to the OS on subsequent
  * major GCs which are NOT caused by a heap overflow.
  *
  * Each major GC which is NOT caused by heap overflow increases the consec_idle_gcs
@@ -2300,3 +2308,81 @@ bool doIdleGCWork(Capability *cap STG_UNUSED, bool all)
  * place, so we always keep that much. If using compacting or nonmoving then we need a lower number,
  * so we just retain at least `1.2 * live_bytes` for some protection.
  */
+
+/* Note [-Fd and thrashing]
+ * ~~~~~~~~~~~~~~~~~~~~~~~~
+ *
+ * See the discussion on #21483 and how low we should scale memory usage and whether
+ * the current behaviour will result in thrashing.
+ *
+ >  The post is correct to say that we need 2L memory to do copying collection
+ >  when there is L live bytes. But it seems to ignore the fact that we don't
+ >  collect until the heap has grown to F*L bytes (where F=2 by default), which
+ >  means that in a steady state of L live bytes we actually need (1+F)L bytes
+ >  (plus the nursery). This is where the original default of returning memory
+ >  above (2+F)L came from: we know we need (1+F)L, plus a bit for the nursery,
+ >  plus we add on another L so that we don't thrash.
+
+ >  If I'm understanding the way -Fd works, it will return memory until we're
+ >  below the 3L value, which will definitely lead to thrashing because we'll
+ >  reallocate the memory to get back to 3L at the next major GC, assuming L
+ >  remains constant.
+
+ >  I think you would want to decay from (2+F)L to (1+F)L, but no lower than
+ >  that. Also, saying that we're "scaling F" is not really the right way to
+ >  think about it, since F is not actually changing, it's the fudge factor we
+ >  want to change.
+
+ * The situation where -Fd kicks in is when the process is idle (ie not
+ * allocating at all), if a process is idle then the memory usage returns down to
+ * a minimal baseline (2L) over quite an extended period by default. Which I
+ * think is what you want because if your program memory usage was increasing,
+ * this is the state you would get into during an idle period. The expectation is
+ * that the amount of program the RTS retains is related to the actual live bytes
+ * rather than the historical memory usage of the program.
+
+ * You are then right, that if after this idle period then we start allocating
+ * again then we will need to get more memory from the OS but this is on the scale
+ * of 10s of minutes rather than milliseconds or seconds (due to the delay
+ * factor).
+ *
+ * It seems in your reply you assume that there will certainly be a major GC due
+ * to allocation in the near future -- but this isn't true. I am often leaving a
+ * high memory consumption Haskell process idle for periods of days/weeks on my
+ * machine and so if that retained 3*L bytes then I would certainly notice!
+ *
+ *
+ * > True enough, I was mainly thinking about programs that allocate continuously.
+ * >
+ * > So let's be a bit more precise
+ * >
+ * >     At the next major GC, the program will need (1+F)L
+ * >     Until the next major GC, it needs from L up to FL bytes
+ * >
+ * > (ignoring the nursery and other things, for simplicity)
+ * >
+ * > if you expect to be in state (2) for a long time, then you could free as
+ * > much memory as you like. Indeed you could free everything except the nursery
+ * > and L after a major GC. Gradually freeing memory instead is a way to say "if
+ * > I've already been idle for a while, then I'm likely to be idle for a while
+ * > longer", which might be true (but it's a hypothesis, like the generational
+ * > hypothesis).
+ * > So if we're hypothesising that the program is idle, why stop freeing at 2L, why not go further?
+ *
+ * Because to my understanding when you are idle you are going to do an idle
+ * major collection which will require 2L so if you free below that you will
+ * get thrashing.
+ *
+ * > But why would you do an idle GC at all in that case? If the program is
+ * > mostly idle, and we think it's going to be mostly idle in the future, and
+ * > we've already free'd all the memory, then there's no reason to do any more
+ * > idle GCs.
+ *
+ * I think it's rare for programs to be completely idle for long periods. Even
+ * during "idle" periods, there is probably still a small amount of allocation
+ * happening before each idle period (hence the reason for flags like -Iw which
+ * prevent many idle collections happening in close succession during periods
+ * of "idleness".
+ *
+
+*/
