@@ -279,8 +279,8 @@ dirty_BLOCKING_QUEUE:
 /* N.B. This happens during the pause so we own all capabilities. */
 void nonmovingSweepMutLists()
 {
-    for (uint32_t n = 0; n < n_capabilities; n++) {
-        Capability *cap = capabilities[n];
+    for (uint32_t n = 0; n < getNumCapabilities(); n++) {
+        Capability *cap = getCapability(n);
         bdescr *old_mut_list = cap->mut_lists[oldest_gen->no];
         cap->mut_lists[oldest_gen->no] = allocBlockOnNode_lock(cap->node);
         for (bdescr *bd = old_mut_list; bd; bd = bd->link) {
