@@ -560,6 +560,19 @@ checkClosure( const StgClosure* p )
         return sizeofW(StgTRecChunk);
       }
 
+    case CONTINUATION:
+    {
+        StgContinuation *cont = (StgContinuation *)p;
+        if (cont->apply_mask_frame) {
+          ASSERT(cont->apply_mask_frame == &stg_unmaskAsyncExceptionszh_ret_info
+              || cont->apply_mask_frame == &stg_maskAsyncExceptionszh_ret_info
+              || cont->apply_mask_frame == &stg_maskUninterruptiblezh_ret_info);
+          ASSERT(LOOKS_LIKE_CLOSURE_PTR(cont->stack + cont->mask_frame_offset));
+        }
+        checkStackChunk(cont->stack, cont->stack + cont->stack_size);
+        return continuation_sizeW(cont);
+    }
+
     default:
         barf("checkClosure (closure type %d)", info->type);
     }
@@ -622,7 +635,7 @@ void checkNonmovingHeap (const struct NonmovingHeap *heap)
         const struct NonmovingAllocator *alloc = heap->allocators[i];
         checkNonmovingSegments(alloc->filled);
         checkNonmovingSegments(alloc->active);
-        for (unsigned int cap=0; cap < n_capabilities; cap++) {
+        for (unsigned int cap=0; cap < getNumCapabilities(); cap++) {
             checkNonmovingSegments(alloc->current[cap]);
         }
     }
@@ -740,6 +753,10 @@ checkTSO(StgTSO *tso)
     ASSERT(LOOKS_LIKE_CLOSURE_PTR(tso->global_link) &&
             (tso->global_link == END_TSO_QUEUE ||
              get_itbl((StgClosure*)tso->global_link)->type == TSO));
+
+    if (tso->label) {
+        ASSERT(LOOKS_LIKE_CLOSURE_PTR(tso->label));
+    }
 }
 
 /*
@@ -818,7 +835,7 @@ checkLocalMutableLists (uint32_t cap_no)
 {
     uint32_t g;
     for (g = 1; g < RtsFlags.GcFlags.generations; g++) {
-        checkMutableList(capabilities[cap_no]->mut_lists[g], g);
+        checkMutableList(getCapability(cap_no)->mut_lists[g], g);
     }
 }
 
@@ -826,7 +843,7 @@ static void
 checkMutableLists (void)
 {
     uint32_t i;
-    for (i = 0; i < n_capabilities; i++) {
+    for (i = 0; i < getNumCapabilities(); i++) {
         checkLocalMutableLists(i);
     }
 }
@@ -945,7 +962,7 @@ static void checkGeneration (generation *gen,
 
     checkHeapChain(gen->blocks);
 
-    for (n = 0; n < n_capabilities; n++) {
+    for (n = 0; n < getNumCapabilities(); n++) {
         ws = &gc_threads[n]->gens[gen->no];
         checkHeapChain(ws->todo_bd);
         checkHeapChain(ws->part_list);
@@ -970,7 +987,7 @@ static void checkFullHeap (bool after_major_gc)
     for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
         checkGeneration(&generations[g], after_major_gc);
     }
-    for (n = 0; n < n_capabilities; n++) {
+    for (n = 0; n < getNumCapabilities(); n++) {
         checkNurserySanity(&nurseries[n]);
     }
 }
@@ -1018,8 +1035,8 @@ findMemoryLeak (void)
 {
     uint32_t g, i, j;
     for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
-        for (i = 0; i < n_capabilities; i++) {
-            markBlocks(capabilities[i]->mut_lists[g]);
+        for (i = 0; i < getNumCapabilities(); i++) {
+            markBlocks(getCapability(i)->mut_lists[g]);
             markBlocks(gc_threads[i]->gens[g].part_list);
             markBlocks(gc_threads[i]->gens[g].scavd_list);
             markBlocks(gc_threads[i]->gens[g].todo_bd);
@@ -1033,10 +1050,10 @@ findMemoryLeak (void)
         markBlocks(nurseries[i].blocks);
     }
 
-    for (i = 0; i < n_capabilities; i++) {
+    for (i = 0; i < getNumCapabilities(); i++) {
         markBlocks(gc_threads[i]->free_blocks);
-        markBlocks(capabilities[i]->pinned_object_block);
-        markBlocks(capabilities[i]->upd_rem_set.queue.blocks);
+        markBlocks(getCapability(i)->pinned_object_block);
+        markBlocks(getCapability(i)->upd_rem_set.queue.blocks);
     }
 
     if (RtsFlags.GcFlags.useNonmoving) {
@@ -1049,7 +1066,7 @@ findMemoryLeak (void)
             struct NonmovingAllocator *alloc = nonmovingHeap.allocators[i];
             markNonMovingSegments(alloc->filled);
             markNonMovingSegments(alloc->active);
-            for (j = 0; j < n_capabilities; j++) {
+            for (j = 0; j < getNumCapabilities(); j++) {
                 markNonMovingSegments(alloc->current[j]);
             }
         }
@@ -1160,7 +1177,7 @@ countNonMovingAllocator(struct NonmovingAllocator *alloc)
 {
     W_ ret = countNonMovingSegments(alloc->filled)
            + countNonMovingSegments(alloc->active);
-    for (uint32_t i = 0; i < n_capabilities; ++i) {
+    for (uint32_t i = 0; i < getNumCapabilities(); ++i) {
         ret += countNonMovingSegments(alloc->current[i]);
     }
     return ret;
@@ -1190,19 +1207,20 @@ memInventory (bool show)
   bool leak;
 
 #if defined(THREADED_RTS)
-  // Can't easily do a memory inventory: We might race with the nonmoving
-  // collector. In principle we could try to take nonmoving_collection_mutex
-  // and do an inventory if we have it but we don't currently implement this.
-  if (RtsFlags.GcFlags.useNonmoving)
-    return;
+  // We need to be careful not to race with the nonmoving collector.
+  // If a nonmoving collection is on-going we simply abort the inventory.
+  if (RtsFlags.GcFlags.useNonmoving){
+    if(TRY_ACQUIRE_LOCK(&nonmoving_collection_mutex))
+      return;
+  }
 #endif
 
   // count the blocks we current have
 
   for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
       gen_blocks[g] = 0;
-      for (i = 0; i < n_capabilities; i++) {
-          gen_blocks[g] += countBlocks(capabilities[i]->mut_lists[g]);
+      for (i = 0; i < getNumCapabilities(); i++) {
+          gen_blocks[g] += countBlocks(getCapability(i)->mut_lists[g]);
           gen_blocks[g] += countBlocks(gc_threads[i]->gens[g].part_list);
           gen_blocks[g] += countBlocks(gc_threads[i]->gens[g].scavd_list);
           gen_blocks[g] += countBlocks(gc_threads[i]->gens[g].todo_bd);
@@ -1214,14 +1232,14 @@ memInventory (bool show)
       ASSERT(countBlocks(nurseries[i].blocks) == nurseries[i].n_blocks);
       nursery_blocks += nurseries[i].n_blocks;
   }
-  for (i = 0; i < n_capabilities; i++) {
+  for (i = 0; i < getNumCapabilities(); i++) {
       W_ n = countBlocks(gc_threads[i]->free_blocks);
       gc_free_blocks += n;
-      if (capabilities[i]->pinned_object_block != NULL) {
-          nursery_blocks += capabilities[i]->pinned_object_block->blocks;
+      if (getCapability(i)->pinned_object_block != NULL) {
+          nursery_blocks += getCapability(i)->pinned_object_block->blocks;
       }
-      nursery_blocks += countBlocks(capabilities[i]->pinned_object_blocks);
-      free_pinned_blocks += countBlocks(capabilities[i]->pinned_object_empty);
+      nursery_blocks += countBlocks(getCapability(i)->pinned_object_blocks);
+      free_pinned_blocks += countBlocks(getCapability(i)->pinned_object_empty);
   }
 
 #if defined(PROFILING)
@@ -1240,8 +1258,8 @@ memInventory (bool show)
   free_blocks = countFreeList();
 
   // count UpdRemSet blocks
-  for (i = 0; i < n_capabilities; ++i) {
-      upd_rem_set_blocks += countBlocks(capabilities[i]->upd_rem_set.queue.blocks);
+  for (i = 0; i < getNumCapabilities(); ++i) {
+      upd_rem_set_blocks += countBlocks(getCapability(i)->upd_rem_set.queue.blocks);
   }
   upd_rem_set_blocks += countBlocks(upd_rem_set_block_list);
 
@@ -1298,6 +1316,13 @@ memInventory (bool show)
   }
   ASSERT(n_alloc_blocks == live_blocks);
   ASSERT(!leak);
+
+#if defined(THREADED_RTS)
+  if (RtsFlags.GcFlags.useNonmoving){
+    RELEASE_LOCK(&nonmoving_collection_mutex);
+  }
+#endif
+
 }
 
 
