@@ -225,7 +225,7 @@ initStorage (void)
 #endif
 
   if (RtsFlags.GcFlags.useNonmoving)
-      nonmovingAddCapabilities(n_capabilities);
+      nonmovingAddCapabilities(getNumCapabilities());
 
   /* The oldest generation has one step. */
   if (RtsFlags.GcFlags.compact || RtsFlags.GcFlags.sweep) {
@@ -257,7 +257,7 @@ initStorage (void)
   for (n = 0; n < n_numa_nodes; n++) {
       next_nursery[n] = n;
   }
-  storageAddCapabilities(0, n_capabilities);
+  storageAddCapabilities(0, getNumCapabilities());
 
   IF_DEBUG(gc, statDescribeGens());
 
@@ -293,8 +293,8 @@ void storageAddCapabilities (uint32_t from, uint32_t to)
     // we've moved the nurseries, so we have to update the rNursery
     // pointers from the Capabilities.
     for (i = 0; i < from; i++) {
-        uint32_t index = capabilities[i]->r.rNursery - old_nurseries;
-        capabilities[i]->r.rNursery = &nurseries[index];
+        uint32_t index = getCapability(i)->r.rNursery - old_nurseries;
+        getCapability(i)->r.rNursery = &nurseries[index];
     }
 
     /* The allocation area.  Policy: keep the allocation area
@@ -316,7 +316,7 @@ void storageAddCapabilities (uint32_t from, uint32_t to)
     // allocate a block for each mut list
     for (n = from; n < to; n++) {
         for (g = 1; g < RtsFlags.GcFlags.generations; g++) {
-            capabilities[n]->mut_lists[g] =
+            getCapability(n)->mut_lists[g] =
                 allocBlockOnNode(capNoToNumaNode(n));
         }
     }
@@ -325,7 +325,7 @@ void storageAddCapabilities (uint32_t from, uint32_t to)
     if (RtsFlags.GcFlags.useNonmoving) {
         nonmovingAddCapabilities(to);
         for (i = 0; i < to; ++i) {
-            init_upd_rem_set(&capabilities[i]->upd_rem_set);
+            nonmovingInitUpdRemSet(&getCapability(i)->upd_rem_set);
         }
     }
 
@@ -375,8 +375,8 @@ void listAllBlocks (ListBlocksCb cb, void *user)
 {
   uint32_t g, i;
   for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
-      for (i = 0; i < n_capabilities; i++) {
-          cb(user, capabilities[i]->mut_lists[g]);
+      for (i = 0; i < getNumCapabilities(); i++) {
+          cb(user, getCapability(i)->mut_lists[g]);
           cb(user, gc_threads[i]->gens[g].part_list);
           cb(user, gc_threads[i]->gens[g].scavd_list);
           cb(user, gc_threads[i]->gens[g].todo_bd);
@@ -387,12 +387,12 @@ void listAllBlocks (ListBlocksCb cb, void *user)
   for (i = 0; i < n_nurseries; i++) {
       cb(user, nurseries[i].blocks);
   }
-  for (i = 0; i < n_capabilities; i++) {
-      if (capabilities[i]->pinned_object_block != NULL) {
-          cb(user, capabilities[i]->pinned_object_block);
+  for (i = 0; i < getNumCapabilities(); i++) {
+      if (getCapability(i)->pinned_object_block != NULL) {
+          cb(user, getCapability(i)->pinned_object_block);
       }
-      cb(user, capabilities[i]->pinned_object_blocks);
-      cb(user, capabilities[i]->pinned_object_empty);
+      cb(user, getCapability(i)->pinned_object_blocks);
+      cb(user, getCapability(i)->pinned_object_empty);
   }
 }
 
@@ -784,8 +784,8 @@ assignNurseriesToCapabilities (uint32_t from, uint32_t to)
     uint32_t i, node;
 
     for (i = from; i < to; i++) {
-        node = capabilities[i]->node;
-        assignNurseryToCapability(capabilities[i], next_nursery[node]);
+        node = getCapability(i)->node;
+        assignNurseryToCapability(getCapability(i), next_nursery[node]);
         next_nursery[node] += n_numa_nodes;
     }
 }
@@ -816,7 +816,7 @@ resetNurseries (void)
     for (n = 0; n < n_numa_nodes; n++) {
         next_nursery[n] = n;
     }
-    assignNurseriesToCapabilities(0, n_capabilities);
+    assignNurseriesToCapabilities(0, getNumCapabilities());
 
 #if defined(DEBUG)
     bdescr *bd;
@@ -1404,11 +1404,14 @@ allocatePinned (Capability *cap, W_ n /*words*/, W_ alignment /*bytes*/, W_ alig
 void
 dirty_MUT_VAR(StgRegTable *reg, StgMutVar *mvar, StgClosure *old)
 {
+#if !defined(THREADED_RTS)
+    // This doesn't hold in the threaded RTS as we may race with another thread.
     ASSERT(RELAXED_LOAD(&mvar->header.info) == &stg_MUT_VAR_CLEAN_info);
+#endif
 
     Capability *cap = regTableToCapability(reg);
     // No barrier required here as no other heap object fields are read. See
-    // note [Heap memory barriers] in SMP.h.
+    // Note [Heap memory barriers] in SMP.h.
     SET_INFO((StgClosure*) mvar, &stg_MUT_VAR_DIRTY_info);
     recordClosureMutated(cap, (StgClosure *) mvar);
     IF_NONMOVING_WRITE_BARRIER_ENABLED {
@@ -1429,7 +1432,7 @@ dirty_TVAR(Capability *cap, StgTVar *p,
            StgClosure *old)
 {
     // No barrier required here as no other heap object fields are read. See
-    // note [Heap memory barriers] in SMP.h.
+    // Note [Heap memory barriers] in SMP.h.
     if (RELAXED_LOAD(&p->header.info) == &stg_TVAR_CLEAN_info) {
         SET_INFO((StgClosure*) p, &stg_TVAR_DIRTY_info);
         recordClosureMutated(cap,(StgClosure*)p);
@@ -1568,12 +1571,12 @@ calcTotalAllocated (void)
     uint64_t tot_alloc = 0;
     W_ n;
 
-    for (n = 0; n < n_capabilities; n++) {
-        tot_alloc += capabilities[n]->total_allocated;
+    for (n = 0; n < getNumCapabilities(); n++) {
+        tot_alloc += getCapability(n)->total_allocated;
 
-        traceEventHeapAllocated(capabilities[n],
+        traceEventHeapAllocated(getCapability(n),
                                 CAPSET_HEAP_DEFAULT,
-                                capabilities[n]->total_allocated * sizeof(W_));
+                                getCapability(n)->total_allocated * sizeof(W_));
     }
 
     return tot_alloc;
@@ -1589,13 +1592,13 @@ updateNurseriesStats (void)
     uint32_t i;
     bdescr *bd;
 
-    for (i = 0; i < n_capabilities; i++) {
+    for (i = 0; i < getNumCapabilities(); i++) {
         // The current nursery block and the current allocate block have not
         // yet been accounted for in cap->total_allocated, so we add them here.
-        bd = capabilities[i]->r.rCurrentNursery;
-        if (bd) finishedNurseryBlock(capabilities[i], bd);
-        bd = capabilities[i]->r.rCurrentAlloc;
-        if (bd) finishedNurseryBlock(capabilities[i], bd);
+        bd = getCapability(i)->r.rCurrentNursery;
+        if (bd) finishedNurseryBlock(getCapability(i), bd);
+        bd = getCapability(i)->r.rCurrentAlloc;
+        if (bd) finishedNurseryBlock(getCapability(i), bd);
     }
 }
 
@@ -1783,6 +1786,7 @@ void flushExec (W_ len, AdjustorExecutable exec_addr)
 #elif (defined(arm_HOST_ARCH) || defined(aarch64_HOST_ARCH)) && (defined(ios_HOST_OS) || defined(darwin_HOST_OS))
   /* On iOS we need to use the special 'sys_icache_invalidate' call. */
   sys_icache_invalidate(exec_addr, len);
+#elif defined(wasm32_HOST_ARCH)
 #elif defined(__clang__)
   unsigned char* begin = (unsigned char*)exec_addr;
   unsigned char* end   = begin + len;
